# (c) 2018, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

''' unit test template for ONTAP Ansible module '''

from __future__ import print_function
import json
import pytest

from units.compat import unittest
from units.compat.mock import patch, Mock
from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes
import ansible.module_utils.netapp as netapp_utils

from ansible.modules.storage.netapp.na_ontap_firewall_policy \
    import NetAppONTAPFirewallPolicy as fp_module  # module under test

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip('skipping as missing required netapp_lib')


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)  # pylint: disable=protected-access


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class MockONTAPConnection(object):
    ''' mock server connection to ONTAP host '''

    def __init__(self, kind=None, data=None):
        ''' save arguments '''
        self.kind = kind
        self.data = data
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        if self.kind == 'policy':
            xml = self.build_policy_info(self.data)
        if self.kind == 'config':
            xml = self.build_firewall_config_info(self.data)
        self.xml_out = xml
        return xml

    @staticmethod
    def build_policy_info(data):
        ''' build xml data for net-firewall-policy-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = {
            'num-records': 1,
            'attributes-list': {
                'net-firewall-policy-info': {
                    'policy': data['policy'],
                    'service': data['service'],
                    'allow-list': [
                        {'ip-and-mask': '1.2.3.0/24'}
                    ]
                }
            }
        }

        xml.translate_struct(attributes)
        return xml

    @staticmethod
    def build_firewall_config_info(data):
        ''' build xml data for net-firewall-config-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = {
            'attributes': {
                'net-firewall-config-info': {
                    'is-enabled': 'true',
                    'is-logging': 'false'
                }
            }
        }
        xml.translate_struct(attributes)
        return xml


class TestMyModule(unittest.TestCase):
    ''' a group of related Unit Tests '''

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)
        self.mock_policy = {
            'policy': 'test',
            'service': 'http',
            'vserver': 'my_vserver',
            'allow_list': '1.2.3.0/24'
        }
        self.mock_config = {
            'node': 'test',
            'enable': 'enable',
            'logging': 'enable'
        }

    def mock_policy_args(self):
        return {
            'policy': self.mock_policy['policy'],
            'service': self.mock_policy['service'],
            'vserver': self.mock_policy['vserver'],
            'allow_list': [self.mock_policy['allow_list']],
            'hostname': 'test',
            'username': 'test_user',
            'password': 'test_pass!'
        }

    def mock_config_args(self):
        return {
            'node': self.mock_config['node'],
            'enable': self.mock_config['enable'],
            'logging': self.mock_config['logging'],
            'hostname': 'test',
            'username': 'test_user',
            'password': 'test_pass!'
        }

    def get_mock_object(self, kind=None):
        """
        Helper method to return an na_ontap_firewall_policy object
        :param kind: passes this param to MockONTAPConnection()
        :return: na_ontap_firewall_policy object
        """
        obj = fp_module()
        obj.autosupport_log = Mock(return_value=None)
        if kind is None:
            obj.server = MockONTAPConnection()
        else:
            mock_data = self.mock_config if kind == 'config' else self.mock_policy
            obj.server = MockONTAPConnection(kind=kind, data=mock_data)
        return obj

    def test_module_fail_when_required_args_missing(self):
        ''' required arguments are reported as errors '''
        with pytest.raises(AnsibleFailJson) as exc:
            set_module_args({})
            fp_module()
        print('Info: %s' % exc.value.args[0]['msg'])

    def test_helper_firewall_policy_attributes(self):
        ''' helper returns dictionary with vserver, service and policy details '''
        data = self.mock_policy
        set_module_args(self.mock_policy_args())
        result = self.get_mock_object('policy').firewall_policy_attributes()
        del data['allow_list']
        assert data == result

    def test_helper_validate_ip_addresses_positive(self):
        ''' test if helper validates if IP is a network address '''
        data = self.mock_policy_args()
        data['allow_list'] = ['1.2.0.0/16', '1.2.3.0/24']
        set_module_args(data)
        result = self.get_mock_object().validate_ip_addresses()
        assert result is None

    def test_helper_validate_ip_addresses_negative(self):
        ''' test if helper validates if IP is a network address '''
        data = self.mock_policy_args()
        data['allow_list'] = ['1.2.0.10/16', '1.2.3.0/24']
        set_module_args(data)
        with pytest.raises(AnsibleFailJson) as exc:
            self.get_mock_object().validate_ip_addresses()
        msg = 'Error: Invalid IP address value for allow_list parameter.' \
              'Please specify a network address without host bits set: ' \
              '1.2.0.10/16 has host bits set'
        assert exc.value.args[0]['msg'] == msg

    def test_get_nonexistent_policy(self):
        ''' Test if get_firewall_policy returns None for non-existent policy '''
        set_module_args(self.mock_policy_args())
        result = self.get_mock_object().get_firewall_policy()
        assert result is None

    def test_get_existing_policy(self):
        ''' Test if get_firewall_policy returns policy details for existing policy '''
        data = self.mock_policy_args()
        set_module_args(data)
        result = self.get_mock_object('policy').get_firewall_policy()
        assert result['service'] == data['service']
        assert result['allow_list'] == ['1.2.3.0/24']  # from build_policy_info()

    def test_successful_create(self):
        ''' Test successful create '''
        set_module_args(self.mock_policy_args())
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object().apply()
        assert exc.value.args[0]['changed']

    def test_create_idempotency(self):
        ''' Test create idempotency '''
        set_module_args(self.mock_policy_args())
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object('policy').apply()
        assert not exc.value.args[0]['changed']

    def test_successful_delete(self):
        ''' Test delete existing job '''
        data = self.mock_policy_args()
        data['state'] = 'absent'
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object('policy').apply()
        assert exc.value.args[0]['changed']

    def test_delete_idempotency(self):
        ''' Test delete idempotency '''
        data = self.mock_policy_args()
        data['state'] = 'absent'
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object().apply()
        assert not exc.value.args[0]['changed']

    def test_successful_modify(self):
        ''' Test successful modify allow_list '''
        data = self.mock_policy_args()
        data['allow_list'] = ['1.2.0.0/16']
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object('policy').apply()
        assert exc.value.args[0]['changed']

    def test_successful_modify_mutiple_ips(self):
        ''' Test successful modify allow_list '''
        data = self.mock_policy_args()
        data['allow_list'] = ['1.2.0.0/16', '1.0.0.0/8']
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object('policy').apply()
        assert exc.value.args[0]['changed']

    def test_get_nonexistent_config(self):
        ''' Test if get_firewall_config returns None for non-existent node '''
        set_module_args(self.mock_config_args())
        result = self.get_mock_object().get_firewall_config_for_node()
        assert result is None

    def test_get_existing_config(self):
        ''' Test if get_firewall_config returns policy details for existing node '''
        data = self.mock_config_args()
        set_module_args(data)
        result = self.get_mock_object('config').get_firewall_config_for_node()
        assert result['enable'] == 'enable'  # from build_config_info()
        assert result['logging'] == 'disable'  # from build_config_info()

    def test_successful_modify_config(self):
        ''' Test successful modify allow_list '''
        data = self.mock_config_args()
        data['enable'] = 'disable'
        data['logging'] = 'enable'
        set_module_args(data)
        with pytest.raises(AnsibleExitJson) as exc:
            self.get_mock_object('config').apply()
        assert exc.value.args[0]['changed']



#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# the lib use python logging can get it if the following is set in your
# Ansible config.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_system_interface
short_description: Configure interfaces in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS by allowing the
      user to set and modify system feature and interface category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.2
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
       description:
            - FortiOS or FortiGate ip address.
       required: true
    username:
        description:
            - FortiOS or FortiGate username.
        required: true
    password:
        description:
            - FortiOS or FortiGate password.
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS
              protocol
        type: bool
        default: true
    system_interface:
        description:
            - Configure interfaces.
        default: null
        suboptions:
            state:
                description:
                    - Indicates whether to create or remove the object
                choices:
                    - present
                    - absent
            ac-name:
                description:
                    - PPPoE server name.
            aggregate:
                description:
                    - Aggregate interface.
            algorithm:
                description:
                    - Frame distribution algorithm.
                choices:
                    - L2
                    - L3
                    - L4
            alias:
                description:
                    - Alias will be displayed with the interface name to make it easier to distinguish.
            allowaccess:
                description:
                    - Permitted types of management access to this interface.
                choices:
                    - ping
                    - https
                    - ssh
                    - snmp
                    - http
                    - telnet
                    - fgfm
                    - radius-acct
                    - probe-response
                    - capwap
                    - ftm
            ap-discover:
                description:
                    - Enable/disable automatic registration of unknown FortiAP devices.
                choices:
                    - enable
                    - disable
            arpforward:
                description:
                    - Enable/disable ARP forwarding.
                choices:
                    - enable
                    - disable
            auth-type:
                description:
                    - PPP authentication type to use.
                choices:
                    - auto
                    - pap
                    - chap
                    - mschapv1
                    - mschapv2
            auto-auth-extension-device:
                description:
                    - Enable/disable automatic authorization of dedicated Fortinet extension device on this interface.
                choices:
                    - enable
                    - disable
            bfd:
                description:
                    - Bidirectional Forwarding Detection (BFD) settings.
                choices:
                    - global
                    - enable
                    - disable
            bfd-desired-min-tx:
                description:
                    - BFD desired minimal transmit interval.
            bfd-detect-mult:
                description:
                    - BFD detection multiplier.
            bfd-required-min-rx:
                description:
                    - BFD required minimal receive interval.
            broadcast-forticlient-discovery:
                description:
                    - Enable/disable broadcasting FortiClient discovery messages.
                choices:
                    - enable
                    - disable
            broadcast-forward:
                description:
                    - Enable/disable broadcast forwarding.
                choices:
                    - enable
                    - disable
            captive-portal:
                description:
                    - Enable/disable captive portal.
            cli-conn-status:
                description:
                    - CLI connection status.
            color:
                description:
                    - Color of icon on the GUI.
            dedicated-to:
                description:
                    - Configure interface for single purpose.
                choices:
                    - none
                    - management
            defaultgw:
                description:
                    - Enable to get the gateway IP from the DHCP or PPPoE server.
                choices:
                    - enable
                    - disable
            description:
                description:
                    - Description.
            detected-peer-mtu:
                description:
                    - MTU of detected peer (0 - 4294967295).
            detectprotocol:
                description:
                    - Protocols used to detect the server.
                choices:
                    - ping
                    - tcp-echo
                    - udp-echo
            detectserver:
                description:
                    - Gateway's ping server for this IP.
            device-access-list:
                description:
                    - Device access list.
            device-identification:
                description:
                    - Enable/disable passively gathering of device identity information about the devices on the network connected to this interface.
                choices:
                    - enable
                    - disable
            device-identification-active-scan:
                description:
                    - Enable/disable active gathering of device identity information about the devices on the network connected to this interface.
                choices:
                    - enable
                    - disable
            device-netscan:
                description:
                    - Enable/disable inclusion of devices detected on this interface in network vulnerability scans.
                choices:
                    - disable
                    - enable
            device-user-identification:
                description:
                    - Enable/disable passive gathering of user identity information about users on this interface.
                choices:
                    - enable
                    - disable
            devindex:
                description:
                    - Device Index.
            dhcp-client-identifier:
                description:
                    - DHCP client identifier.
            dhcp-relay-agent-option:
                description:
                    - Enable/disable DHCP relay agent option.
                choices:
                    - enable
                    - disable
            dhcp-relay-ip:
                description:
                    - DHCP relay IP address.
            dhcp-relay-service:
                description:
                    - Enable/disable allowing this interface to act as a DHCP relay.
                choices:
                    - disable
                    - enable
            dhcp-relay-type:
                description:
                    - DHCP relay type (regular or IPsec).
                choices:
                    - regular
                    - ipsec
            dhcp-renew-time:
                description:
                    - DHCP renew time in seconds (300-604800), 0 means use the renew time provided by the server.
            disc-retry-timeout:
                description:
                    - Time in seconds to wait before retrying to start a PPPoE discovery, 0 means no timeout.
            disconnect-threshold:
                description:
                    - Time in milliseconds to wait before sending a notification that this interface is down or disconnected.
            distance:
                description:
                    - Distance for routes learned through PPPoE or DHCP, lower distance indicates preferred route.
            dns-server-override:
                description:
                    - Enable/disable use DNS acquired by DHCP or PPPoE.
                choices:
                    - enable
                    - disable
            drop-fragment:
                description:
                    - Enable/disable drop fragment packets.
                choices:
                    - enable
                    - disable
            drop-overlapped-fragment:
                description:
                    - Enable/disable drop overlapped fragment packets.
                choices:
                    - enable
                    - disable
            egress-shaping-profile:
                description:
                    - Outgoing traffic shaping profile.
            endpoint-compliance:
                description:
                    - Enable/disable endpoint compliance enforcement.
                choices:
                    - enable
                    - disable
            estimated-downstream-bandwidth:
                description:
                    - Estimated maximum downstream bandwidth (kbps). Used to estimate link utilization.
            estimated-upstream-bandwidth:
                description:
                    - Estimated maximum upstream bandwidth (kbps). Used to estimate link utilization.
            explicit-ftp-proxy:
                description:
                    - Enable/disable the explicit FTP proxy on this interface.
                choices:
                    - enable
                    - disable
            explicit-web-proxy:
                description:
                    - Enable/disable the explicit web proxy on this interface.
                choices:
                    - enable
                    - disable
            external:
                description:
                    - Enable/disable identifying the interface as an external interface (which usually means it's connected to the Internet).
                choices:
                    - enable
                    - disable
            fail-action-on-extender:
                description:
                    - Action on extender when interface fail .
                choices:
                    - soft-restart
                    - hard-restart
                    - reboot
            fail-alert-interfaces:
                description:
                    - Names of the FortiGate interfaces from which the link failure alert is sent for this interface.
                suboptions:
                    name:
                        description:
                            - Names of the physical interfaces belonging to the aggregate or redundant interface. Source system.interface.name.
                        required: true
            fail-alert-method:
                description:
                    - Select link-failed-signal or link-down method to alert about a failed link.
                choices:
                    - link-failed-signal
                    - link-down
            fail-detect:
                description:
                    - Enable/disable fail detection features for this interface.
                choices:
                    - enable
                    - disable
            fail-detect-option:
                description:
                    - Options for detecting that this interface has failed.
                choices:
                    - detectserver
                    - link-down
            fortiheartbeat:
                description:
                    - Enable/disable FortiHeartBeat (FortiTelemetry on GUI).
                choices:
                    - enable
                    - disable
            fortilink:
                description:
                    - Enable FortiLink to dedicate this interface to manage other Fortinet devices.
                choices:
                    - enable
                    - disable
            fortilink-backup-link:
                description:
                    - fortilink split interface backup link.
            fortilink-split-interface:
                description:
                    - Enable/disable FortiLink split interface to connect member link to different FortiSwitch in stack for uplink redundancy (maximum 2
                       interfaces in the "members" command).
                choices:
                    - enable
                    - disable
            fortilink-stacking:
                description:
                    - Enable/disable FortiLink switch-stacking on this interface.
                choices:
                    - enable
                    - disable
            forward-domain:
                description:
                    - Transparent mode forward domain.
            gwdetect:
                description:
                    - Enable/disable detect gateway alive for first.
                choices:
                    - enable
                    - disable
            ha-priority:
                description:
                    - HA election priority for the PING server.
            icmp-accept-redirect:
                description:
                    - Enable/disable ICMP accept redirect.
                choices:
                    - enable
                    - disable
            icmp-send-redirect:
                description:
                    - Enable/disable ICMP send redirect.
                choices:
                    - enable
                    - disable
            ident-accept:
                description:
                    - Enable/disable authentication for this interface.
                choices:
                    - enable
                    - disable
            idle-timeout:
                description:
                    - PPPoE auto disconnect after idle timeout seconds, 0 means no timeout.
            inbandwidth:
                description:
                    - Bandwidth limit for incoming traffic (0 - 16776000 kbps), 0 means unlimited.
            ingress-spillover-threshold:
                description:
                    - Ingress Spillover threshold (0 - 16776000 kbps).
            interface:
                description:
                    - Interface name. Source system.interface.name.
            internal:
                description:
                    - Implicitly created.
            ip:
                description:
                    - "Interface IPv4 address and subnet mask, syntax: X.X.X.X/24."
            ipmac:
                description:
                    - Enable/disable IP/MAC binding.
                choices:
                    - enable
                    - disable
            ips-sniffer-mode:
                description:
                    - Enable/disable the use of this interface as a one-armed sniffer.
                choices:
                    - enable
                    - disable
            ipunnumbered:
                description:
                    - Unnumbered IP used for PPPoE interfaces for which no unique local address is provided.
            ipv6:
                description:
                    - IPv6 of interface.
                suboptions:
                    autoconf:
                        description:
                            - Enable/disable address auto config.
                        choices:
                            - enable
                            - disable
                    dhcp6-client-options:
                        description:
                            - DHCPv6 client options.
                        choices:
                            - rapid
                            - iapd
                            - iana
                    dhcp6-information-request:
                        description:
                            - Enable/disable DHCPv6 information request.
                        choices:
                            - enable
                            - disable
                    dhcp6-prefix-delegation:
                        description:
                            - Enable/disable DHCPv6 prefix delegation.
                        choices:
                            - enable
                            - disable
                    dhcp6-prefix-hint:
                        description:
                            - DHCPv6 prefix that will be used as a hint to the upstream DHCPv6 server.
                    dhcp6-prefix-hint-plt:
                        description:
                            - DHCPv6 prefix hint preferred life time (sec), 0 means unlimited lease time.
                    dhcp6-prefix-hint-vlt:
                        description:
                            - DHCPv6 prefix hint valid life time (sec).
                    dhcp6-relay-ip:
                        description:
                            - DHCPv6 relay IP address.
                    dhcp6-relay-service:
                        description:
                            - Enable/disable DHCPv6 relay.
                        choices:
                            - disable
                            - enable
                    dhcp6-relay-type:
                        description:
                            - DHCPv6 relay type.
                        choices:
                            - regular
                    ip6-address:
                        description:
                            - "Primary IPv6 address prefix, syntax: xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx/xxx"
                    ip6-allowaccess:
                        description:
                            - Allow management access to the interface.
                        choices:
                            - ping
                            - https
                            - ssh
                            - snmp
                            - http
                            - telnet
                            - fgfm
                            - capwap
                    ip6-default-life:
                        description:
                            - Default life (sec).
                    ip6-delegated-prefix-list:
                        description:
                            - Advertised IPv6 delegated prefix list.
                        suboptions:
                            autonomous-flag:
                                description:
                                    - Enable/disable the autonomous flag.
                                choices:
                                    - enable
                                    - disable
                            onlink-flag:
                                description:
                                    - Enable/disable the onlink flag.
                                choices:
                                    - enable
                                    - disable
                            prefix-id:
                                description:
                                    - Prefix ID.
                                required: true
                            rdnss:
                                description:
                                    - Recursive DNS server option.
                            rdnss-service:
                                description:
                                    - Recursive DNS service option.
                                choices:
                                    - delegated
                                    - default
                                    - specify
                            subnet:
                                description:
                                    -  Add subnet ID to routing prefix.
                            upstream-interface:
                                description:
                                    - Name of the interface that provides delegated information. Source system.interface.name.
                    ip6-dns-server-override:
                        description:
                            - Enable/disable using the DNS server acquired by DHCP.
                        choices:
                            - enable
                            - disable
                    ip6-extra-addr:
                        description:
                            - Extra IPv6 address prefixes of interface.
                        suboptions:
                            prefix:
                                description:
                                    - IPv6 address prefix.
                                required: true
                    ip6-hop-limit:
                        description:
                            - Hop limit (0 means unspecified).
                    ip6-link-mtu:
                        description:
                            - IPv6 link MTU.
                    ip6-manage-flag:
                        description:
                            - Enable/disable the managed flag.
                        choices:
                            - enable
                            - disable
                    ip6-max-interval:
                        description:
                            - IPv6 maximum interval (4 to 1800 sec).
                    ip6-min-interval:
                        description:
                            - IPv6 minimum interval (3 to 1350 sec).
                    ip6-mode:
                        description:
                            - Addressing mode (static, DHCP, delegated).
                        choices:
                            - static
                            - dhcp
                            - pppoe
                            - delegated
                    ip6-other-flag:
                        description:
                            - Enable/disable the other IPv6 flag.
                        choices:
                            - enable
                            - disable
                    ip6-prefix-list:
                        description:
                            - Advertised prefix list.
                        suboptions:
                            autonomous-flag:
                                description:
                                    - Enable/disable the autonomous flag.
                                choices:
                                    - enable
                                    - disable
                            dnssl:
                                description:
                                    - DNS search list option.
                                suboptions:
                                    domain:
                                        description:
                                            - Domain name.
                                        required: true
                            onlink-flag:
                                description:
                                    - Enable/disable the onlink flag.
                                choices:
                                    - enable
                                    - disable
                            preferred-life-time:
                                description:
                                    - Preferred life time (sec).
                            prefix:
                                description:
                                    - IPv6 prefix.
                                required: true
                            rdnss:
                                description:
                                    - Recursive DNS server option.
                            valid-life-time:
                                description:
                                    - Valid life time (sec).
                    ip6-reachable-time:
                        description:
                            - IPv6 reachable time (milliseconds; 0 means unspecified).
                    ip6-retrans-time:
                        description:
                            - IPv6 retransmit time (milliseconds; 0 means unspecified).
                    ip6-send-adv:
                        description:
                            - Enable/disable sending advertisements about the interface.
                        choices:
                            - enable
                            - disable
                    ip6-subnet:
                        description:
                            - " Subnet to routing prefix, syntax: xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx/xxx"
                    ip6-upstream-interface:
                        description:
                            - Interface name providing delegated information. Source system.interface.name.
                    nd-cert:
                        description:
                            - Neighbor discovery certificate. Source certificate.local.name.
                    nd-cga-modifier:
                        description:
                            - Neighbor discovery CGA modifier.
                    nd-mode:
                        description:
                            - Neighbor discovery mode.
                        choices:
                            - basic
                            - SEND-compatible
                    nd-security-level:
                        description:
                            - Neighbor discovery security level (0 - 7; 0 = least secure, default = 0).
                    nd-timestamp-delta:
                        description:
                            - Neighbor discovery timestamp delta value (1 - 3600 sec; default = 300).
                    nd-timestamp-fuzz:
                        description:
                            - Neighbor discovery timestamp fuzz factor (1 - 60 sec; default = 1).
                    vrip6_link_local:
                        description:
                            - Link-local IPv6 address of virtual router.
                    vrrp-virtual-mac6:
                        description:
                            - Enable/disable virtual MAC for VRRP.
                        choices:
                            - enable
                            - disable
                    vrrp6:
                        description:
                            - IPv6 VRRP configuration.
                        suboptions:
                            accept-mode:
                                description:
                                    - Enable/disable accept mode.
                                choices:
                                    - enable
                                    - disable
                            adv-interval:
                                description:
                                    - Advertisement interval (1 - 255 seconds).
                            preempt:
                                description:
                                    - Enable/disable preempt mode.
                                choices:
                                    - enable
                                    - disable
                            priority:
                                description:
                                    - Priority of the virtual router (1 - 255).
                            start-time:
                                description:
                                    - Startup time (1 - 255 seconds).
                            status:
                                description:
                                    - Enable/disable VRRP.
                                choices:
                                    - enable
                                    - disable
                            vrdst6:
                                description:
                                    - Monitor the route to this destination.
                            vrgrp:
                                description:
                                    - VRRP group ID (1 - 65535).
                            vrid:
                                description:
                                    - Virtual router identifier (1 - 255).
                                required: true
                            vrip6:
                                description:
                                    - IPv6 address of the virtual router.
            l2forward:
                description:
                    - Enable/disable l2 forwarding.
                choices:
                    - enable
                    - disable
            lacp-ha-slave:
                description:
                    - LACP HA slave.
                choices:
                    - enable
                    - disable
            lacp-mode:
                description:
                    - LACP mode.
                choices:
                    - static
                    - passive
                    - active
            lacp-speed:
                description:
                    - How often the interface sends LACP messages.
                choices:
                    - slow
                    - fast
            lcp-echo-interval:
                description:
                    - Time in seconds between PPPoE Link Control Protocol (LCP) echo requests.
            lcp-max-echo-fails:
                description:
                    - Maximum missed LCP echo messages before disconnect.
            link-up-delay:
                description:
                    - Number of milliseconds to wait before considering a link is up.
            lldp-transmission:
                description:
                    - Enable/disable Link Layer Discovery Protocol (LLDP) transmission.
                choices:
                    - enable
                    - disable
                    - vdom
            macaddr:
                description:
                    - Change the interface's MAC address.
            managed-device:
                description:
                    - Available when FortiLink is enabled, used for managed devices through FortiLink interface.
                suboptions:
                    name:
                        description:
                            - Managed dev identifier.
                        required: true
            management-ip:
                description:
                    - High Availability in-band management IP address of this interface.
            member:
                description:
                    - Physical interfaces that belong to the aggregate or redundant interface.
                suboptions:
                    interface-name:
                        description:
                            - Physical interface name. Source system.interface.name.
                        required: true
            min-links:
                description:
                    - Minimum number of aggregated ports that must be up.
            min-links-down:
                description:
                    - Action to take when less than the configured minimum number of links are active.
                choices:
                    - operational
                    - administrative
            mode:
                description:
                    - Addressing mode (static, DHCP, PPPoE).
                choices:
                    - static
                    - dhcp
                    - pppoe
            mtu:
                description:
                    - MTU value for this interface.
            mtu-override:
                description:
                    - Enable to set a custom MTU for this interface.
                choices:
                    - enable
                    - disable
            name:
                description:
                    - Name.
                required: true
            ndiscforward:
                description:
                    - Enable/disable NDISC forwarding.
                choices:
                    - enable
                    - disable
            netbios-forward:
                description:
                    - Enable/disable NETBIOS forwarding.
                choices:
                    - disable
                    - enable
            netflow-sampler:
                description:
                    - Enable/disable NetFlow on this interface and set the data that NetFlow collects (rx, tx, or both).
                choices:
                    - disable
                    - tx
                    - rx
                    - both
            outbandwidth:
                description:
                    - Bandwidth limit for outgoing traffic (0 - 16776000 kbps).
            padt-retry-timeout:
                description:
                    - PPPoE Active Discovery Terminate (PADT) used to terminate sessions after an idle time.
            password:
                description:
                    - PPPoE account's password.
            ping-serv-status:
                description:
                    - PING server status.
            polling-interval:
                description:
                    - sFlow polling interval (1 - 255 sec).
            pppoe-unnumbered-negotiate:
                description:
                    - Enable/disable PPPoE unnumbered negotiation.
                choices:
                    - enable
                    - disable
            pptp-auth-type:
                description:
                    - PPTP authentication type.
                choices:
                    - auto
                    - pap
                    - chap
                    - mschapv1
                    - mschapv2
            pptp-client:
                description:
                    - Enable/disable PPTP client.
                choices:
                    - enable
                    - disable
            pptp-password:
                description:
                    - PPTP password.
            pptp-server-ip:
                description:
                    - PPTP server IP address.
            pptp-timeout:
                description:
                    - Idle timer in minutes (0 for disabled).
            pptp-user:
                description:
                    - PPTP user name.
            preserve-session-route:
                description:
                    - Enable/disable preservation of session route when dirty.
                choices:
                    - enable
                    - disable
            priority:
                description:
                    - Priority of learned routes.
            priority-override:
                description:
                    - Enable/disable fail back to higher priority port once recovered.
                choices:
                    - enable
                    - disable
            proxy-captive-portal:
                description:
                    - Enable/disable proxy captive portal on this interface.
                choices:
                    - enable
                    - disable
            redundant-interface:
                description:
                    - Redundant interface.
            remote-ip:
                description:
                    - Remote IP address of tunnel.
            replacemsg-override-group:
                description:
                    - Replacement message override group.
            role:
                description:
                    - Interface role.
                choices:
                    - lan
                    - wan
                    - dmz
                    - undefined
            sample-direction:
                description:
                    - Data that NetFlow collects (rx, tx, or both).
                choices:
                    - tx
                    - rx
                    - both
            sample-rate:
                description:
                    - sFlow sample rate (10 - 99999).
            scan-botnet-connections:
                description:
                    - Enable monitoring or blocking connections to Botnet servers through this interface.
                choices:
                    - disable
                    - block
                    - monitor
            secondary-IP:
                description:
                    - Enable/disable adding a secondary IP to this interface.
                choices:
                    - enable
                    - disable
            secondaryip:
                description:
                    - Second IP address of interface.
                suboptions:
                    allowaccess:
                        description:
                            - Management access settings for the secondary IP address.
                        choices:
                            - ping
                            - https
                            - ssh
                            - snmp
                            - http
                            - telnet
                            - fgfm
                            - radius-acct
                            - probe-response
                            - capwap
                            - ftm
                    detectprotocol:
                        description:
                            - Protocols used to detect the server.
                        choices:
                            - ping
                            - tcp-echo
                            - udp-echo
                    detectserver:
                        description:
                            - Gateway's ping server for this IP.
                    gwdetect:
                        description:
                            - Enable/disable detect gateway alive for first.
                        choices:
                            - enable
                            - disable
                    ha-priority:
                        description:
                            - HA election priority for the PING server.
                    id:
                        description:
                            - ID.
                        required: true
                    ip:
                        description:
                            - Secondary IP address of the interface.
                    ping-serv-status:
                        description:
                            - PING server status.
            security-exempt-list:
                description:
                    - Name of security-exempt-list.
            security-external-logout:
                description:
                    - URL of external authentication logout server.
            security-external-web:
                description:
                    - URL of external authentication web server.
            security-groups:
                description:
                    - User groups that can authenticate with the captive portal.
                suboptions:
                    name:
                        description:
                            - Names of user groups that can authenticate with the captive portal.
                        required: true
            security-mac-auth-bypass:
                description:
                    - Enable/disable MAC authentication bypass.
                choices:
                    - enable
                    - disable
            security-mode:
                description:
                    - Turn on captive portal authentication for this interface.
                choices:
                    - none
                    - captive-portal
                    - 802.1X
            security-redirect-url:
                description:
                    - URL redirection after disclaimer/authentication.
            service-name:
                description:
                    - PPPoE service name.
            sflow-sampler:
                description:
                    - Enable/disable sFlow on this interface.
                choices:
                    - enable
                    - disable
            snmp-index:
                description:
                    - Permanent SNMP Index of the interface.
            speed:
                description:
                    - Interface speed. The default setting and the options available depend on the interface hardware.
                choices:
                    - auto
                    - 10full
                    - 10half
                    - 100full
                    - 100half
                    - 1000full
                    - 1000half
                    - 1000auto
            spillover-threshold:
                description:
                    - Egress Spillover threshold (0 - 16776000 kbps), 0 means unlimited.
            src-check:
                description:
                    - Enable/disable source IP check.
                choices:
                    - enable
                    - disable
            status:
                description:
                    - Bring the interface up or shut the interface down.
                choices:
                    - up
                    - down
            stpforward:
                description:
                    - Enable/disable STP forwarding.
                choices:
                    - enable
                    - disable
            stpforward-mode:
                description:
                    - Configure STP forwarding mode.
                choices:
                    - rpl-all-ext-id
                    - rpl-bridge-ext-id
                    - rpl-nothing
            subst:
                description:
                    - Enable to always send packets from this interface to a destination MAC address.
                choices:
                    - enable
                    - disable
            substitute-dst-mac:
                description:
                    - Destination MAC address that all packets are sent to from this interface.
            switch:
                description:
                    - Contained in switch.
            switch-controller-access-vlan:
                description:
                    - Block FortiSwitch port-to-port traffic.
                choices:
                    - enable
                    - disable
            switch-controller-arp-inspection:
                description:
                    - Enable/disable FortiSwitch ARP inspection.
                choices:
                    - enable
                    - disable
            switch-controller-dhcp-snooping:
                description:
                    - Switch controller DHCP snooping.
                choices:
                    - enable
                    - disable
            switch-controller-dhcp-snooping-option82:
                description:
                    - Switch controller DHCP snooping option82.
                choices:
                    - enable
                    - disable
            switch-controller-dhcp-snooping-verify-mac:
                description:
                    - Switch controller DHCP snooping verify MAC.
                choices:
                    - enable
                    - disable
            switch-controller-igmp-snooping:
                description:
                    - Switch controller IGMP snooping.
                choices:
                    - enable
                    - disable
            switch-controller-learning-limit:
                description:
                    - Limit the number of dynamic MAC addresses on this VLAN (1 - 128, 0 = no limit, default).
            tagging:
                description:
                    - Config object tagging.
                suboptions:
                    category:
                        description:
                            - Tag category. Source system.object-tagging.category.
                    name:
                        description:
                            - Tagging entry name.
                        required: true
                    tags:
                        description:
                            - Tags.
                        suboptions:
                            name:
                                description:
                                    - Tag name. Source system.object-tagging.tags.name.
                                required: true
            tcp-mss:
                description:
                    - TCP maximum segment size. 0 means do not change segment size.
            trust-ip-1:
                description:
                    - Trusted host for dedicated management traffic (0.0.0.0/24 for all hosts).
            trust-ip-2:
                description:
                    - Trusted host for dedicated management traffic (0.0.0.0/24 for all hosts).
            trust-ip-3:
                description:
                    - Trusted host for dedicated management traffic (0.0.0.0/24 for all hosts).
            trust-ip6-1:
                description:
                    - "Trusted IPv6 host for dedicated management traffic (::/0 for all hosts)."
            trust-ip6-2:
                description:
                    - "Trusted IPv6 host for dedicated management traffic (::/0 for all hosts)."
            trust-ip6-3:
                description:
                    - "Trusted IPv6 host for dedicated management traffic (::/0 for all hosts)."
            type:
                description:
                    - Interface type.
                choices:
                    - physical
                    - vlan
                    - aggregate
                    - redundant
                    - tunnel
                    - vdom-link
                    - loopback
                    - switch
                    - hard-switch
                    - vap-switch
                    - wl-mesh
                    - fext-wan
                    - vxlan
                    - hdlc
                    - switch-vlan
            username:
                description:
                    - Username of the PPPoE account, provided by your ISP.
            vdom:
                description:
                    - Interface is in this virtual domain (VDOM). Source system.vdom.name.
            vindex:
                description:
                    - Switch control interface VLAN ID.
            vlanforward:
                description:
                    - Enable/disable traffic forwarding between VLANs on this interface.
                choices:
                    - enable
                    - disable
            vlanid:
                description:
                    - VLAN ID (1 - 4094).
            vrf:
                description:
                    - Virtual Routing Forwarding ID.
            vrrp:
                description:
                    - VRRP configuration.
                suboptions:
                    accept-mode:
                        description:
                            - Enable/disable accept mode.
                        choices:
                            - enable
                            - disable
                    adv-interval:
                        description:
                            - Advertisement interval (1 - 255 seconds).
                    preempt:
                        description:
                            - Enable/disable preempt mode.
                        choices:
                            - enable
                            - disable
                    priority:
                        description:
                            - Priority of the virtual router (1 - 255).
                    proxy-arp:
                        description:
                            - VRRP Proxy ARP configuration.
                        suboptions:
                            id:
                                description:
                                    - ID.
                                required: true
                            ip:
                                description:
                                    - Set IP addresses of proxy ARP.
                    start-time:
                        description:
                            - Startup time (1 - 255 seconds).
                    status:
                        description:
                            - Enable/disable this VRRP configuration.
                        choices:
                            - enable
                            - disable
                    version:
                        description:
                            - VRRP version.
                        choices:
                            - 2
                            - 3
                    vrdst:
                        description:
                            - Monitor the route to this destination.
                    vrdst-priority:
                        description:
                            - Priority of the virtual router when the virtual router destination becomes unreachable (0 - 254).
                    vrgrp:
                        description:
                            - VRRP group ID (1 - 65535).
                    vrid:
                        description:
                            - Virtual router identifier (1 - 255).
                        required: true
                    vrip:
                        description:
                            - IP address of the virtual router.
            vrrp-virtual-mac:
                description:
                    - Enable/disable use of virtual MAC for VRRP.
                choices:
                    - enable
                    - disable
            wccp:
                description:
                    - Enable/disable WCCP on this interface. Used for encapsulated WCCP communication between WCCP clients and servers.
                choices:
                    - enable
                    - disable
            weight:
                description:
                    - Default weight for static routes (if route has no weight configured).
            wins-ip:
                description:
                    - WINS server IP.
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
  tasks:
  - name: Configure interfaces.
    fortios_system_interface:
      host:  "{{ host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{ vdom }}"
      https: "False"
      system_interface:
        state: "present"
        ac-name: "<your_own_value>"
        aggregate: "<your_own_value>"
        algorithm: "L2"
        alias: "<your_own_value>"
        allowaccess: "ping"
        ap-discover: "enable"
        arpforward: "enable"
        auth-type: "auto"
        auto-auth-extension-device: "enable"
        bfd: "global"
        bfd-desired-min-tx: "13"
        bfd-detect-mult: "14"
        bfd-required-min-rx: "15"
        broadcast-forticlient-discovery: "enable"
        broadcast-forward: "enable"
        captive-portal: "18"
        cli-conn-status: "19"
        color: "20"
        dedicated-to: "none"
        defaultgw: "enable"
        description: "<your_own_value>"
        detected-peer-mtu: "24"
        detectprotocol: "ping"
        detectserver: "<your_own_value>"
        device-access-list: "<your_own_value>"
        device-identification: "enable"
        device-identification-active-scan: "enable"
        device-netscan: "disable"
        device-user-identification: "enable"
        devindex: "32"
        dhcp-client-identifier:  "myId_33"
        dhcp-relay-agent-option: "enable"
        dhcp-relay-ip: "<your_own_value>"
        dhcp-relay-service: "disable"
        dhcp-relay-type: "regular"
        dhcp-renew-time: "38"
        disc-retry-timeout: "39"
        disconnect-threshold: "40"
        distance: "41"
        dns-server-override: "enable"
        drop-fragment: "enable"
        drop-overlapped-fragment: "enable"
        egress-shaping-profile: "<your_own_value>"
        endpoint-compliance: "enable"
        estimated-downstream-bandwidth: "47"
        estimated-upstream-bandwidth: "48"
        explicit-ftp-proxy: "enable"
        explicit-web-proxy: "enable"
        external: "enable"
        fail-action-on-extender: "soft-restart"
        fail-alert-interfaces:
         -
            name: "default_name_54 (source system.interface.name)"
        fail-alert-method: "link-failed-signal"
        fail-detect: "enable"
        fail-detect-option: "detectserver"
        fortiheartbeat: "enable"
        fortilink: "enable"
        fortilink-backup-link: "60"
        fortilink-split-interface: "enable"
        fortilink-stacking: "enable"
        forward-domain: "63"
        gwdetect: "enable"
        ha-priority: "65"
        icmp-accept-redirect: "enable"
        icmp-send-redirect: "enable"
        ident-accept: "enable"
        idle-timeout: "69"
        inbandwidth: "70"
        ingress-spillover-threshold: "71"
        interface: "<your_own_value> (source system.interface.name)"
        internal: "73"
        ip: "<your_own_value>"
        ipmac: "enable"
        ips-sniffer-mode: "enable"
        ipunnumbered: "<your_own_value>"
        ipv6:
            autoconf: "enable"
            dhcp6-client-options: "rapid"
            dhcp6-information-request: "enable"
            dhcp6-prefix-delegation: "enable"
            dhcp6-prefix-hint: "<your_own_value>"
            dhcp6-prefix-hint-plt: "84"
            dhcp6-prefix-hint-vlt: "85"
            dhcp6-relay-ip: "<your_own_value>"
            dhcp6-relay-service: "disable"
            dhcp6-relay-type: "regular"
            ip6-address: "<your_own_value>"
            ip6-allowaccess: "ping"
            ip6-default-life: "91"
            ip6-delegated-prefix-list:
             -
                autonomous-flag: "enable"
                onlink-flag: "enable"
                prefix-id: "95"
                rdnss: "<your_own_value>"
                rdnss-service: "delegated"
                subnet: "<your_own_value>"
                upstream-interface: "<your_own_value> (source system.interface.name)"
            ip6-dns-server-override: "enable"
            ip6-extra-addr:
             -
                prefix: "<your_own_value>"
            ip6-hop-limit: "103"
            ip6-link-mtu: "104"
            ip6-manage-flag: "enable"
            ip6-max-interval: "106"
            ip6-min-interval: "107"
            ip6-mode: "static"
            ip6-other-flag: "enable"
            ip6-prefix-list:
             -
                autonomous-flag: "enable"
                dnssl:
                 -
                    domain: "<your_own_value>"
                onlink-flag: "enable"
                preferred-life-time: "115"
                prefix: "<your_own_value>"
                rdnss: "<your_own_value>"
                valid-life-time: "118"
            ip6-reachable-time: "119"
            ip6-retrans-time: "120"
            ip6-send-adv: "enable"
            ip6-subnet: "<your_own_value>"
            ip6-upstream-interface: "<your_own_value> (source system.interface.name)"
            nd-cert: "<your_own_value> (source certificate.local.name)"
            nd-cga-modifier: "<your_own_value>"
            nd-mode: "basic"
            nd-security-level: "127"
            nd-timestamp-delta: "128"
            nd-timestamp-fuzz: "129"
            vrip6_link_local: "<your_own_value>"
            vrrp-virtual-mac6: "enable"
            vrrp6:
             -
                accept-mode: "enable"
                adv-interval: "134"
                preempt: "enable"
                priority: "136"
                start-time: "137"
                status: "enable"
                vrdst6: "<your_own_value>"
                vrgrp: "140"
                vrid: "141"
                vrip6: "<your_own_value>"
        l2forward: "enable"
        lacp-ha-slave: "enable"
        lacp-mode: "static"
        lacp-speed: "slow"
        lcp-echo-interval: "147"
        lcp-max-echo-fails: "148"
        link-up-delay: "149"
        lldp-transmission: "enable"
        macaddr: "<your_own_value>"
        managed-device:
         -
            name: "default_name_153"
        management-ip: "<your_own_value>"
        member:
         -
            interface-name: "<your_own_value> (source system.interface.name)"
        min-links: "157"
        min-links-down: "operational"
        mode: "static"
        mtu: "160"
        mtu-override: "enable"
        name: "default_name_162"
        ndiscforward: "enable"
        netbios-forward: "disable"
        netflow-sampler: "disable"
        outbandwidth: "166"
        padt-retry-timeout: "167"
        password: "<your_own_value>"
        ping-serv-status: "169"
        polling-interval: "170"
        pppoe-unnumbered-negotiate: "enable"
        pptp-auth-type: "auto"
        pptp-client: "enable"
        pptp-password: "<your_own_value>"
        pptp-server-ip: "<your_own_value>"
        pptp-timeout: "176"
        pptp-user: "<your_own_value>"
        preserve-session-route: "enable"
        priority: "179"
        priority-override: "enable"
        proxy-captive-portal: "enable"
        redundant-interface: "<your_own_value>"
        remote-ip: "<your_own_value>"
        replacemsg-override-group: "<your_own_value>"
        role: "lan"
        sample-direction: "tx"
        sample-rate: "187"
        scan-botnet-connections: "disable"
        secondary-IP: "enable"
        secondaryip:
         -
            allowaccess: "ping"
            detectprotocol: "ping"
            detectserver: "<your_own_value>"
            gwdetect: "enable"
            ha-priority: "195"
            id:  "196"
            ip: "<your_own_value>"
            ping-serv-status: "198"
        security-exempt-list: "<your_own_value>"
        security-external-logout: "<your_own_value>"
        security-external-web: "<your_own_value>"
        security-groups:
         -
            name: "default_name_203"
        security-mac-auth-bypass: "enable"
        security-mode: "none"
        security-redirect-url: "<your_own_value>"
        service-name: "<your_own_value>"
        sflow-sampler: "enable"
        snmp-index: "209"
        speed: "auto"
        spillover-threshold: "211"
        src-check: "enable"
        status: "up"
        stpforward: "enable"
        stpforward-mode: "rpl-all-ext-id"
        subst: "enable"
        substitute-dst-mac: "<your_own_value>"
        switch: "<your_own_value>"
        switch-controller-access-vlan: "enable"
        switch-controller-arp-inspection: "enable"
        switch-controller-dhcp-snooping: "enable"
        switch-controller-dhcp-snooping-option82: "enable"
        switch-controller-dhcp-snooping-verify-mac: "enable"
        switch-controller-igmp-snooping: "enable"
        switch-controller-learning-limit: "225"
        tagging:
         -
            category: "<your_own_value> (source system.object-tagging.category)"
            name: "default_name_228"
            tags:
             -
                name: "default_name_230 (source system.object-tagging.tags.name)"
        tcp-mss: "231"
        trust-ip-1: "<your_own_value>"
        trust-ip-2: "<your_own_value>"
        trust-ip-3: "<your_own_value>"
        trust-ip6-1: "<your_own_value>"
        trust-ip6-2: "<your_own_value>"
        trust-ip6-3: "<your_own_value>"
        type: "physical"
        username: "<your_own_value>"
        vdom: "<your_own_value> (source system.vdom.name)"
        vindex: "241"
        vlanforward: "enable"
        vlanid: "243"
        vrf: "244"
        vrrp:
         -
            accept-mode: "enable"
            adv-interval: "247"
            preempt: "enable"
            priority: "249"
            proxy-arp:
             -
                id:  "251"
                ip: "<your_own_value>"
            start-time: "253"
            status: "enable"
            version: "2"
            vrdst: "<your_own_value>"
            vrdst-priority: "257"
            vrgrp: "258"
            vrid: "259"
            vrip: "<your_own_value>"
        vrrp-virtual-mac: "enable"
        wccp: "enable"
        weight: "263"
        wins-ip: "<your_own_value>"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule

fos = None


def login(data):
    host = data['host']
    username = data['username']
    password = data['password']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password)


def filter_system_interface_data(json):
    option_list = ['ac-name', 'aggregate', 'algorithm',
                   'alias', 'allowaccess', 'ap-discover',
                   'arpforward', 'auth-type', 'auto-auth-extension-device',
                   'bfd', 'bfd-desired-min-tx', 'bfd-detect-mult',
                   'bfd-required-min-rx', 'broadcast-forticlient-discovery', 'broadcast-forward',
                   'captive-portal', 'cli-conn-status', 'color',
                   'dedicated-to', 'defaultgw', 'description',
                   'detected-peer-mtu', 'detectprotocol', 'detectserver',
                   'device-access-list', 'device-identification', 'device-identification-active-scan',
                   'device-netscan', 'device-user-identification', 'devindex',
                   'dhcp-client-identifier', 'dhcp-relay-agent-option', 'dhcp-relay-ip',
                   'dhcp-relay-service', 'dhcp-relay-type', 'dhcp-renew-time',
                   'disc-retry-timeout', 'disconnect-threshold', 'distance',
                   'dns-server-override', 'drop-fragment', 'drop-overlapped-fragment',
                   'egress-shaping-profile', 'endpoint-compliance', 'estimated-downstream-bandwidth',
                   'estimated-upstream-bandwidth', 'explicit-ftp-proxy', 'explicit-web-proxy',
                   'external', 'fail-action-on-extender', 'fail-alert-interfaces',
                   'fail-alert-method', 'fail-detect', 'fail-detect-option',
                   'fortiheartbeat', 'fortilink', 'fortilink-backup-link',
                   'fortilink-split-interface', 'fortilink-stacking', 'forward-domain',
                   'gwdetect', 'ha-priority', 'icmp-accept-redirect',
                   'icmp-send-redirect', 'ident-accept', 'idle-timeout',
                   'inbandwidth', 'ingress-spillover-threshold', 'interface',
                   'internal', 'ip', 'ipmac',
                   'ips-sniffer-mode', 'ipunnumbered', 'ipv6',
                   'l2forward', 'lacp-ha-slave', 'lacp-mode',
                   'lacp-speed', 'lcp-echo-interval', 'lcp-max-echo-fails',
                   'link-up-delay', 'lldp-transmission', 'macaddr',
                   'managed-device', 'management-ip', 'member',
                   'min-links', 'min-links-down', 'mode',
                   'mtu', 'mtu-override', 'name',
                   'ndiscforward', 'netbios-forward', 'netflow-sampler',
                   'outbandwidth', 'padt-retry-timeout', 'password',
                   'ping-serv-status', 'polling-interval', 'pppoe-unnumbered-negotiate',
                   'pptp-auth-type', 'pptp-client', 'pptp-password',
                   'pptp-server-ip', 'pptp-timeout', 'pptp-user',
                   'preserve-session-route', 'priority', 'priority-override',
                   'proxy-captive-portal', 'redundant-interface', 'remote-ip',
                   'replacemsg-override-group', 'role', 'sample-direction',
                   'sample-rate', 'scan-botnet-connections', 'secondary-IP',
                   'secondaryip', 'security-exempt-list', 'security-external-logout',
                   'security-external-web', 'security-groups', 'security-mac-auth-bypass',
                   'security-mode', 'security-redirect-url', 'service-name',
                   'sflow-sampler', 'snmp-index', 'speed',
                   'spillover-threshold', 'src-check', 'status',
                   'stpforward', 'stpforward-mode', 'subst',
                   'substitute-dst-mac', 'switch', 'switch-controller-access-vlan',
                   'switch-controller-arp-inspection', 'switch-controller-dhcp-snooping', 'switch-controller-dhcp-snooping-option82',
                   'switch-controller-dhcp-snooping-verify-mac', 'switch-controller-igmp-snooping', 'switch-controller-learning-limit',
                   'tagging', 'tcp-mss', 'trust-ip-1',
                   'trust-ip-2', 'trust-ip-3', 'trust-ip6-1',
                   'trust-ip6-2', 'trust-ip6-3', 'type',
                   'username', 'vdom', 'vindex',
                   'vlanforward', 'vlanid', 'vrf',
                   'vrrp', 'vrrp-virtual-mac', 'wccp',
                   'weight', 'wins-ip']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def flatten_multilists_attributes(data):
    multilist_attrs = [[u'allowaccess'], [u'ipv6', u'ip6-allowaccess']]

    for attr in multilist_attrs:
        try:
            path = "data['" + "']['".join(elem for elem in attr) + "']"
            current_val = eval(path)
            flattened_val = ' '.join(elem for elem in current_val)
            exec(path + '= flattened_val')
        except BaseException:
            pass

    return data


def system_interface(data, fos):
    vdom = data['vdom']
    system_interface_data = data['system_interface']
    flattened_data = flatten_multilists_attributes(system_interface_data)
    filtered_data = filter_system_interface_data(flattened_data)
    if system_interface_data['state'] == "present":
        return fos.set('system',
                       'interface',
                       data=filtered_data,
                       vdom=vdom)

    elif system_interface_data['state'] == "absent":
        return fos.delete('system',
                          'interface',
                          mkey=filtered_data['name'],
                          vdom=vdom)


def fortios_system(data, fos):
    login(data)

    if data['system_interface']:
        resp = system_interface(data, fos)

    fos.logout()
    return not resp['status'] == "success", resp['status'] == "success", resp


def main():
    fields = {
        "host": {"required": True, "type": "str"},
        "username": {"required": True, "type": "str"},
        "password": {"required": False, "type": "str", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "system_interface": {
            "required": False, "type": "dict",
            "options": {
                "state": {"required": True, "type": "str",
                          "choices": ["present", "absent"]},
                "ac-name": {"required": False, "type": "str"},
                "aggregate": {"required": False, "type": "str"},
                "algorithm": {"required": False, "type": "str",
                              "choices": ["L2", "L3", "L4"]},
                "alias": {"required": False, "type": "str"},
                "allowaccess": {"required": False, "type": "list",
                                "choices": ["ping", "https", "ssh",
                                            "snmp", "http", "telnet",
                                            "fgfm", "radius-acct", "probe-response",
                                            "capwap", "ftm"]},
                "ap-discover": {"required": False, "type": "str",
                                "choices": ["enable", "disable"]},
                "arpforward": {"required": False, "type": "str",
                               "choices": ["enable", "disable"]},
                "auth-type": {"required": False, "type": "str",
                              "choices": ["auto", "pap", "chap",
                                          "mschapv1", "mschapv2"]},
                "auto-auth-extension-device": {"required": False, "type": "str",
                                               "choices": ["enable", "disable"]},
                "bfd": {"required": False, "type": "str",
                        "choices": ["global", "enable", "disable"]},
                "bfd-desired-min-tx": {"required": False, "type": "int"},
                "bfd-detect-mult": {"required": False, "type": "int"},
                "bfd-required-min-rx": {"required": False, "type": "int"},
                "broadcast-forticlient-discovery": {"required": False, "type": "str",
                                                    "choices": ["enable", "disable"]},
                "broadcast-forward": {"required": False, "type": "str",
                                      "choices": ["enable", "disable"]},
                "captive-portal": {"required": False, "type": "int"},
                "cli-conn-status": {"required": False, "type": "int"},
                "color": {"required": False, "type": "int"},
                "dedicated-to": {"required": False, "type": "str",
                                 "choices": ["none", "management"]},
                "defaultgw": {"required": False, "type": "str",
                              "choices": ["enable", "disable"]},
                "description": {"required": False, "type": "str"},
                "detected-peer-mtu": {"required": False, "type": "int"},
                "detectprotocol": {"required": False, "type": "str",
                                   "choices": ["ping", "tcp-echo", "udp-echo"]},
                "detectserver": {"required": False, "type": "str"},
                "device-access-list": {"required": False, "type": "str"},
                "device-identification": {"required": False, "type": "str",
                                          "choices": ["enable", "disable"]},
                "device-identification-active-scan": {"required": False, "type": "str",
                                                      "choices": ["enable", "disable"]},
                "device-netscan": {"required": False, "type": "str",
                                   "choices": ["disable", "enable"]},
                "device-user-identification": {"required": False, "type": "str",
                                               "choices": ["enable", "disable"]},
                "devindex": {"required": False, "type": "int"},
                "dhcp-client-identifier": {"required": False, "type": "str"},
                "dhcp-relay-agent-option": {"required": False, "type": "str",
                                            "choices": ["enable", "disable"]},
                "dhcp-relay-ip": {"required": False, "type": "str"},
                "dhcp-relay-service": {"required": False, "type": "str",
                                       "choices": ["disable", "enable"]},
                "dhcp-relay-type": {"required": False, "type": "str",
                                    "choices": ["regular", "ipsec"]},
                "dhcp-renew-time": {"required": False, "type": "int"},
                "disc-retry-timeout": {"required": False, "type": "int"},
                "disconnect-threshold": {"required": False, "type": "int"},
                "distance": {"required": False, "type": "int"},
                "dns-server-override": {"required": False, "type": "str",
                                        "choices": ["enable", "disable"]},
                "drop-fragment": {"required": False, "type": "str",
                                  "choices": ["enable", "disable"]},
                "drop-overlapped-fragment": {"required": False, "type": "str",
                                             "choices": ["enable", "disable"]},
                "egress-shaping-profile": {"required": False, "type": "str"},
                "endpoint-compliance": {"required": False, "type": "str",
                                        "choices": ["enable", "disable"]},
                "estimated-downstream-bandwidth": {"required": False, "type": "int"},
                "estimated-upstream-bandwidth": {"required": False, "type": "int"},
                "explicit-ftp-proxy": {"required": False, "type": "str",
                                       "choices": ["enable", "disable"]},
                "explicit-web-proxy": {"required": False, "type": "str",
                                       "choices": ["enable", "disable"]},
                "external": {"required": False, "type": "str",
                             "choices": ["enable", "disable"]},
                "fail-action-on-extender": {"required": False, "type": "str",
                                            "choices": ["soft-restart", "hard-restart", "reboot"]},
                "fail-alert-interfaces": {"required": False, "type": "list",
                                          "options": {
                                              "name": {"required": True, "type": "str"}
                                          }},
                "fail-alert-method": {"required": False, "type": "str",
                                      "choices": ["link-failed-signal", "link-down"]},
                "fail-detect": {"required": False, "type": "str",
                                "choices": ["enable", "disable"]},
                "fail-detect-option": {"required": False, "type": "str",
                                       "choices": ["detectserver", "link-down"]},
                "fortiheartbeat": {"required": False, "type": "str",
                                   "choices": ["enable", "disable"]},
                "fortilink": {"required": False, "type": "str",
                              "choices": ["enable", "disable"]},
                "fortilink-backup-link": {"required": False, "type": "int"},
                "fortilink-split-interface": {"required": False, "type": "str",
                                              "choices": ["enable", "disable"]},
                "fortilink-stacking": {"required": False, "type": "str",
                                       "choices": ["enable", "disable"]},
                "forward-domain": {"required": False, "type": "int"},
                "gwdetect": {"required": False, "type": "str",
                             "choices": ["enable", "disable"]},
                "ha-priority": {"required": False, "type": "int"},
                "icmp-accept-redirect": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                "icmp-send-redirect": {"required": False, "type": "str",
                                       "choices": ["enable", "disable"]},
                "ident-accept": {"required": False, "type": "str",
                                 "choices": ["enable", "disable"]},
                "idle-timeout": {"required": False, "type": "int"},
                "inbandwidth": {"required": False, "type": "int"},
                "ingress-spillover-threshold": {"required": False, "type": "int"},
                "interface": {"required": False, "type": "str"},
                "internal": {"required": False, "type": "int"},
                "ip": {"required": False, "type": "str"},
                "ipmac": {"required": False, "type": "str",
                          "choices": ["enable", "disable"]},
                "ips-sniffer-mode": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "ipunnumbered": {"required": False, "type": "str"},
                "ipv6": {"required": False, "type": "dict",
                         "options": {
                             "autoconf": {"required": False, "type": "str",
                                          "choices": ["enable", "disable"]},
                             "dhcp6-client-options": {"required": False, "type": "str",
                                                      "choices": ["rapid", "iapd", "iana"]},
                             "dhcp6-information-request": {"required": False, "type": "str",
                                                           "choices": ["enable", "disable"]},
                             "dhcp6-prefix-delegation": {"required": False, "type": "str",
                                                         "choices": ["enable", "disable"]},
                             "dhcp6-prefix-hint": {"required": False, "type": "str"},
                             "dhcp6-prefix-hint-plt": {"required": False, "type": "int"},
                             "dhcp6-prefix-hint-vlt": {"required": False, "type": "int"},
                             "dhcp6-relay-ip": {"required": False, "type": "str"},
                             "dhcp6-relay-service": {"required": False, "type": "str",
                                                     "choices": ["disable", "enable"]},
                             "dhcp6-relay-type": {"required": False, "type": "str",
                                                  "choices": ["regular"]},
                             "ip6-address": {"required": False, "type": "str"},
                             "ip6-allowaccess": {"required": False, "type": "list",
                                                 "choices": ["ping", "https", "ssh",
                                                             "snmp", "http", "telnet",
                                                             "fgfm", "capwap"]},
                             "ip6-default-life": {"required": False, "type": "int"},
                             "ip6-delegated-prefix-list": {"required": False, "type": "list",
                                                           "options": {
                                                               "autonomous-flag": {"required": False, "type": "str",
                                                                                   "choices": ["enable", "disable"]},
                                                               "onlink-flag": {"required": False, "type": "str",
                                                                               "choices": ["enable", "disable"]},
                                                               "prefix-id": {"required": True, "type": "int"},
                                                               "rdnss": {"required": False, "type": "str"},
                                                               "rdnss-service": {"required": False, "type": "str",
                                                                                 "choices": ["delegated", "default", "specify"]},
                                                               "subnet": {"required": False, "type": "str"},
                                                               "upstream-interface": {"required": False, "type": "str"}
                                                           }},
                             "ip6-dns-server-override": {"required": False, "type": "str",
                                                         "choices": ["enable", "disable"]},
                             "ip6-extra-addr": {"required": False, "type": "list",
                                                "options": {
                                                    "prefix": {"required": True, "type": "str"}
                                                }},
                             "ip6-hop-limit": {"required": False, "type": "int"},
                             "ip6-link-mtu": {"required": False, "type": "int"},
                             "ip6-manage-flag": {"required": False, "type": "str",
                                                 "choices": ["enable", "disable"]},
                             "ip6-max-interval": {"required": False, "type": "int"},
                             "ip6-min-interval": {"required": False, "type": "int"},
                             "ip6-mode": {"required": False, "type": "str",
                                          "choices": ["static", "dhcp", "pppoe",
                                                      "delegated"]},
                             "ip6-other-flag": {"required": False, "type": "str",
                                                "choices": ["enable", "disable"]},
                             "ip6-prefix-list": {"required": False, "type": "list",
                                                 "options": {
                                                     "autonomous-flag": {"required": False, "type": "str",
                                                                         "choices": ["enable", "disable"]},
                                                     "dnssl": {"required": False, "type": "list",
                                                               "options": {
                                                                   "domain": {"required": True, "type": "str"}
                                                               }},
                                                     "onlink-flag": {"required": False, "type": "str",
                                                                     "choices": ["enable", "disable"]},
                                                     "preferred-life-time": {"required": False, "type": "int"},
                                                     "prefix": {"required": True, "type": "str"},
                                                     "rdnss": {"required": False, "type": "str"},
                                                     "valid-life-time": {"required": False, "type": "int"}
                                                 }},
                             "ip6-reachable-time": {"required": False, "type": "int"},
                             "ip6-retrans-time": {"required": False, "type": "int"},
                             "ip6-send-adv": {"required": False, "type": "str",
                                              "choices": ["enable", "disable"]},
                             "ip6-subnet": {"required": False, "type": "str"},
                             "ip6-upstream-interface": {"required": False, "type": "str"},
                             "nd-cert": {"required": False, "type": "str"},
                             "nd-cga-modifier": {"required": False, "type": "str"},
                             "nd-mode": {"required": False, "type": "str",
                                         "choices": ["basic", "SEND-compatible"]},
                             "nd-security-level": {"required": False, "type": "int"},
                             "nd-timestamp-delta": {"required": False, "type": "int"},
                             "nd-timestamp-fuzz": {"required": False, "type": "int"},
                             "vrip6_link_local": {"required": False, "type": "str"},
                             "vrrp-virtual-mac6": {"required": False, "type": "str",
                                                   "choices": ["enable", "disable"]},
                             "vrrp6": {"required": False, "type": "list",
                                       "options": {
                                           "accept-mode": {"required": False, "type": "str",
                                                           "choices": ["enable", "disable"]},
                                           "adv-interval": {"required": False, "type": "int"},
                                           "preempt": {"required": False, "type": "str",
                                                       "choices": ["enable", "disable"]},
                                           "priority": {"required": False, "type": "int"},
                                           "start-time": {"required": False, "type": "int"},
                                           "status": {"required": False, "type": "str",
                                                      "choices": ["enable", "disable"]},
                                           "vrdst6": {"required": False, "type": "str"},
                                           "vrgrp": {"required": False, "type": "int"},
                                           "vrid": {"required": True, "type": "int"},
                                           "vrip6": {"required": False, "type": "str"}
                                       }}
                         }},
                "l2forward": {"required": False, "type": "str",
                              "choices": ["enable", "disable"]},
                "lacp-ha-slave": {"required": False, "type": "str",
                                  "choices": ["enable", "disable"]},
                "lacp-mode": {"required": False, "type": "str",
                              "choices": ["static", "passive", "active"]},
                "lacp-speed": {"required": False, "type": "str",
                               "choices": ["slow", "fast"]},
                "lcp-echo-interval": {"required": False, "type": "int"},
                "lcp-max-echo-fails": {"required": False, "type": "int"},
                "link-up-delay": {"required": False, "type": "int"},
                "lldp-transmission": {"required": False, "type": "str",
                                      "choices": ["enable", "disable", "vdom"]},
                "macaddr": {"required": False, "type": "str"},
                "managed-device": {"required": False, "type": "list",
                                   "options": {
                                       "name": {"required": True, "type": "str"}
                                   }},
                "management-ip": {"required": False, "type": "str"},
                "member": {"required": False, "type": "list",
                           "options": {
                               "interface-name": {"required": True, "type": "str"}
                           }},
                "min-links": {"required": False, "type": "int"},
                "min-links-down": {"required": False, "type": "str",
                                   "choices": ["operational", "administrative"]},
                "mode": {"required": False, "type": "str",
                         "choices": ["static", "dhcp", "pppoe"]},
                "mtu": {"required": False, "type": "int"},
                "mtu-override": {"required": False, "type": "str",
                                 "choices": ["enable", "disable"]},
                "name": {"required": True, "type": "str"},
                "ndiscforward": {"required": False, "type": "str",
                                 "choices": ["enable", "disable"]},
                "netbios-forward": {"required": False, "type": "str",
                                    "choices": ["disable", "enable"]},
                "netflow-sampler": {"required": False, "type": "str",
                                    "choices": ["disable", "tx", "rx",
                                                "both"]},
                "outbandwidth": {"required": False, "type": "int"},
                "padt-retry-timeout": {"required": False, "type": "int"},
                "password": {"required": False, "type": "str"},
                "ping-serv-status": {"required": False, "type": "int"},
                "polling-interval": {"required": False, "type": "int"},
                "pppoe-unnumbered-negotiate": {"required": False, "type": "str",
                                               "choices": ["enable", "disable"]},
                "pptp-auth-type": {"required": False, "type": "str",
                                   "choices": ["auto", "pap", "chap",
                                               "mschapv1", "mschapv2"]},
                "pptp-client": {"required": False, "type": "str",
                                "choices": ["enable", "disable"]},
                "pptp-password": {"required": False, "type": "str"},
                "pptp-server-ip": {"required": False, "type": "str"},
                "pptp-timeout": {"required": False, "type": "int"},
                "pptp-user": {"required": False, "type": "str"},
                "preserve-session-route": {"required": False, "type": "str",
                                           "choices": ["enable", "disable"]},
                "priority": {"required": False, "type": "int"},
                "priority-override": {"required": False, "type": "str",
                                      "choices": ["enable", "disable"]},
                "proxy-captive-portal": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                "redundant-interface": {"required": False, "type": "str"},
                "remote-ip": {"required": False, "type": "str"},
                "replacemsg-override-group": {"required": False, "type": "str"},
                "role": {"required": False, "type": "str",
                         "choices": ["lan", "wan", "dmz",
                                     "undefined"]},
                "sample-direction": {"required": False, "type": "str",
                                     "choices": ["tx", "rx", "both"]},
                "sample-rate": {"required": False, "type": "int"},
                "scan-botnet-connections": {"required": False, "type": "str",
                                            "choices": ["disable", "block", "monitor"]},
                "secondary-IP": {"required": False, "type": "str",
                                 "choices": ["enable", "disable"]},
                "secondaryip": {"required": False, "type": "list",
                                "options": {
                                    "allowaccess": {"required": False, "type": "str",
                                                    "choices": ["ping", "https", "ssh",
                                                                "snmp", "http", "telnet",
                                                                "fgfm", "radius-acct", "probe-response",
                                                                "capwap", "ftm"]},
                                    "detectprotocol": {"required": False, "type": "str",
                                                       "choices": ["ping", "tcp-echo", "udp-echo"]},
                                    "detectserver": {"required": False, "type": "str"},
                                    "gwdetect": {"required": False, "type": "str",
                                                 "choices": ["enable", "disable"]},
                                    "ha-priority": {"required": False, "type": "int"},
                                    "id": {"required": True, "type": "int"},
                                    "ip": {"required": False, "type": "str"},
                                    "ping-serv-status": {"required": False, "type": "int"}
                                }},
                "security-exempt-list": {"required": False, "type": "str"},
                "security-external-logout": {"required": False, "type": "str"},
                "security-external-web": {"required": False, "type": "str"},
                "security-groups": {"required": False, "type": "list",
                                    "options": {
                                        "name": {"required": True, "type": "str"}
                                    }},
                "security-mac-auth-bypass": {"required": False, "type": "str",
                                             "choices": ["enable", "disable"]},
                "security-mode": {"required": False, "type": "str",
                                  "choices": ["none", "captive-portal", "802.1X"]},
                "security-redirect-url": {"required": False, "type": "str"},
                "service-name": {"required": False, "type": "str"},
                "sflow-sampler": {"required": False, "type": "str",
                                  "choices": ["enable", "disable"]},
                "snmp-index": {"required": False, "type": "int"},
                "speed": {"required": False, "type": "str",
                          "choices": ["auto", "10full", "10half",
                                      "100full", "100half", "1000full",
                                      "1000half", "1000auto"]},
                "spillover-threshold": {"required": False, "type": "int"},
                "src-check": {"required": False, "type": "str",
                              "choices": ["enable", "disable"]},
                "status": {"required": False, "type": "str",
                           "choices": ["up", "down"]},
                "stpforward": {"required": False, "type": "str",
                               "choices": ["enable", "disable"]},
                "stpforward-mode": {"required": False, "type": "str",
                                    "choices": ["rpl-all-ext-id", "rpl-bridge-ext-id", "rpl-nothing"]},
                "subst": {"required": False, "type": "str",
                          "choices": ["enable", "disable"]},
                "substitute-dst-mac": {"required": False, "type": "str"},
                "switch": {"required": False, "type": "str"},
                "switch-controller-access-vlan": {"required": False, "type": "str",
                                                  "choices": ["enable", "disable"]},
                "switch-controller-arp-inspection": {"required": False, "type": "str",
                                                     "choices": ["enable", "disable"]},
                "switch-controller-dhcp-snooping": {"required": False, "type": "str",
                                                    "choices": ["enable", "disable"]},
                "switch-controller-dhcp-snooping-option82": {"required": False, "type": "str",
                                                             "choices": ["enable", "disable"]},
                "switch-controller-dhcp-snooping-verify-mac": {"required": False, "type": "str",
                                                               "choices": ["enable", "disable"]},
                "switch-controller-igmp-snooping": {"required": False, "type": "str",
                                                    "choices": ["enable", "disable"]},
                "switch-controller-learning-limit": {"required": False, "type": "int"},
                "tagging": {"required": False, "type": "list",
                            "options": {
                                "category": {"required": False, "type": "str"},
                                "name": {"required": True, "type": "str"},
                                "tags": {"required": False, "type": "list",
                                         "options": {
                                             "name": {"required": True, "type": "str"}
                                         }}
                            }},
                "tcp-mss": {"required": False, "type": "int"},
                "trust-ip-1": {"required": False, "type": "str"},
                "trust-ip-2": {"required": False, "type": "str"},
                "trust-ip-3": {"required": False, "type": "str"},
                "trust-ip6-1": {"required": False, "type": "str"},
                "trust-ip6-2": {"required": False, "type": "str"},
                "trust-ip6-3": {"required": False, "type": "str"},
                "type": {"required": False, "type": "str",
                         "choices": ["physical", "vlan", "aggregate",
                                     "redundant", "tunnel", "vdom-link",
                                     "loopback", "switch", "hard-switch",
                                     "vap-switch", "wl-mesh", "fext-wan",
                                     "vxlan", "hdlc", "switch-vlan"]},
                "username": {"required": False, "type": "str"},
                "vdom": {"required": False, "type": "str"},
                "vindex": {"required": False, "type": "int"},
                "vlanforward": {"required": False, "type": "str",
                                "choices": ["enable", "disable"]},
                "vlanid": {"required": False, "type": "int"},
                "vrf": {"required": False, "type": "int"},
                "vrrp": {"required": False, "type": "list",
                         "options": {
                             "accept-mode": {"required": False, "type": "str",
                                             "choices": ["enable", "disable"]},
                             "adv-interval": {"required": False, "type": "int"},
                             "preempt": {"required": False, "type": "str",
                                         "choices": ["enable", "disable"]},
                             "priority": {"required": False, "type": "int"},
                             "proxy-arp": {"required": False, "type": "list",
                                           "options": {
                                               "id": {"required": True, "type": "int"},
                                               "ip": {"required": False, "type": "str"}
                                           }},
                             "start-time": {"required": False, "type": "int"},
                             "status": {"required": False, "type": "str",
                                        "choices": ["enable", "disable"]},
                             "version": {"required": False, "type": "str",
                                         "choices": ["2", "3"]},
                             "vrdst": {"required": False, "type": "str"},
                             "vrdst-priority": {"required": False, "type": "int"},
                             "vrgrp": {"required": False, "type": "int"},
                             "vrid": {"required": True, "type": "int"},
                             "vrip": {"required": False, "type": "str"}
                         }},
                "vrrp-virtual-mac": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                "wccp": {"required": False, "type": "str",
                         "choices": ["enable", "disable"]},
                "weight": {"required": False, "type": "int"},
                "wins-ip": {"required": False, "type": "str"}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)
    try:
        from fortiosapi import FortiOSAPI
    except ImportError:
        module.fail_json(msg="fortiosapi module is required")

    global fos
    fos = FortiOSAPI()

    is_error, has_changed, result = fortios_system(module.params, fos)

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Copyright (c), Ted Timmons <ted@timmons.me>, 2017.
# Most of this was originally added by other creators in the postgresql_user module.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

try:
    import psycopg2
    import psycopg2.extras
    HAS_PSYCOPG2 = True
except ImportError:
    HAS_PSYCOPG2 = False

from ansible.module_utils._text import to_native


class LibraryError(Exception):
    pass


def ensure_libs(sslrootcert=None):
    if not HAS_PSYCOPG2:
        raise LibraryError('psycopg2 is not installed. we need psycopg2.')
    if sslrootcert and psycopg2.__version__ < '2.4.3':
        raise LibraryError('psycopg2 must be at least 2.4.3 in order to use the ca_cert parameter')

    # no problems
    return None


def postgres_common_argument_spec():
    return dict(
        login_user=dict(default='postgres'),
        login_password=dict(default='', no_log=True),
        login_host=dict(default=''),
        login_unix_socket=dict(default=''),
        port=dict(type='int', default=5432, aliases=['login_port']),
        ssl_mode=dict(default='prefer', choices=['allow', 'disable', 'prefer', 'require', 'verify-ca', 'verify-full']),
        ca_cert=dict(aliases=['ssl_rootcert']),
    )


def connect_to_db(module, kw, autocommit=False):
    try:
        db_connection = psycopg2.connect(**kw)
        if autocommit:
            if psycopg2.__version__ >= '2.4.2':
                db_connection.set_session(autocommit=True)
            else:
                db_connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)

    except TypeError as e:
        if 'sslrootcert' in e.args[0]:
            module.fail_json(msg='Postgresql server must be at least '
                                 'version 8.4 to support sslrootcert')

        module.fail_json(msg="unable to connect to database: %s" % to_native(e))

    except Exception as e:
        module.fail_json(msg="unable to connect to database: %s" % to_native(e))

    return db_connection


def get_pg_version(cursor):
    cursor.execute("select current_setting('server_version_num')")
    return int(cursor.fetchone()[0])



#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_waf_profile
short_description: Web application firewall configuration in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS by allowing the
      user to set and modify waf feature and profile category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.2
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
       description:
            - FortiOS or FortiGate ip address.
       required: true
    username:
        description:
            - FortiOS or FortiGate username.
        required: true
    password:
        description:
            - FortiOS or FortiGate password.
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS
              protocol
        type: bool
        default: true
    waf_profile:
        description:
            - Web application firewall configuration.
        default: null
        suboptions:
            state:
                description:
                    - Indicates whether to create or remove the object
                choices:
                    - present
                    - absent
            address-list:
                description:
                    - Black address list and white address list.
                suboptions:
                    blocked-address:
                        description:
                            - Blocked address.
                        suboptions:
                            name:
                                description:
                                    - Address name. Source firewall.address.name firewall.addrgrp.name.
                                required: true
                    blocked-log:
                        description:
                            - Enable/disable logging on blocked addresses.
                        choices:
                            - enable
                            - disable
                    severity:
                        description:
                            - Severity.
                        choices:
                            - high
                            - medium
                            - low
                    status:
                        description:
                            - Status.
                        choices:
                            - enable
                            - disable
                    trusted-address:
                        description:
                            - Trusted address.
                        suboptions:
                            name:
                                description:
                                    - Address name. Source firewall.address.name firewall.addrgrp.name.
                                required: true
            comment:
                description:
                    - Comment.
            constraint:
                description:
                    - WAF HTTP protocol restrictions.
                suboptions:
                    content-length:
                        description:
                            - HTTP content length in request.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            length:
                                description:
                                    - Length of HTTP content in bytes (0 to 2147483647).
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    exception:
                        description:
                            - HTTP constraint exception.
                        suboptions:
                            address:
                                description:
                                    - Host address. Source firewall.address.name firewall.addrgrp.name.
                            content-length:
                                description:
                                    - HTTP content length in request.
                                choices:
                                    - enable
                                    - disable
                            header-length:
                                description:
                                    - HTTP header length in request.
                                choices:
                                    - enable
                                    - disable
                            hostname:
                                description:
                                    - Enable/disable hostname check.
                                choices:
                                    - enable
                                    - disable
                            id:
                                description:
                                    - Exception ID.
                                required: true
                            line-length:
                                description:
                                    - HTTP line length in request.
                                choices:
                                    - enable
                                    - disable
                            malformed:
                                description:
                                    - Enable/disable malformed HTTP request check.
                                choices:
                                    - enable
                                    - disable
                            max-cookie:
                                description:
                                    - Maximum number of cookies in HTTP request.
                                choices:
                                    - enable
                                    - disable
                            max-header-line:
                                description:
                                    - Maximum number of HTTP header line.
                                choices:
                                    - enable
                                    - disable
                            max-range-segment:
                                description:
                                    - Maximum number of range segments in HTTP range line.
                                choices:
                                    - enable
                                    - disable
                            max-url-param:
                                description:
                                    - Maximum number of parameters in URL.
                                choices:
                                    - enable
                                    - disable
                            method:
                                description:
                                    - Enable/disable HTTP method check.
                                choices:
                                    - enable
                                    - disable
                            param-length:
                                description:
                                    - Maximum length of parameter in URL, HTTP POST request or HTTP body.
                                choices:
                                    - enable
                                    - disable
                            pattern:
                                description:
                                    - URL pattern.
                            regex:
                                description:
                                    - Enable/disable regular expression based pattern match.
                                choices:
                                    - enable
                                    - disable
                            url-param-length:
                                description:
                                    - Maximum length of parameter in URL.
                                choices:
                                    - enable
                                    - disable
                            version:
                                description:
                                    - Enable/disable HTTP version check.
                                choices:
                                    - enable
                                    - disable
                    header-length:
                        description:
                            - HTTP header length in request.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            length:
                                description:
                                    - Length of HTTP header in bytes (0 to 2147483647).
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    hostname:
                        description:
                            - Enable/disable hostname check.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    line-length:
                        description:
                            - HTTP line length in request.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            length:
                                description:
                                    - Length of HTTP line in bytes (0 to 2147483647).
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    malformed:
                        description:
                            - Enable/disable malformed HTTP request check.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    max-cookie:
                        description:
                            - Maximum number of cookies in HTTP request.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            max-cookie:
                                description:
                                    - Maximum number of cookies in HTTP request (0 to 2147483647).
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    max-header-line:
                        description:
                            - Maximum number of HTTP header line.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            max-header-line:
                                description:
                                    - Maximum number HTTP header lines (0 to 2147483647).
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    max-range-segment:
                        description:
                            - Maximum number of range segments in HTTP range line.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            max-range-segment:
                                description:
                                    - Maximum number of range segments in HTTP range line (0 to 2147483647).
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    max-url-param:
                        description:
                            - Maximum number of parameters in URL.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            max-url-param:
                                description:
                                    - Maximum number of parameters in URL (0 to 2147483647).
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    method:
                        description:
                            - Enable/disable HTTP method check.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    param-length:
                        description:
                            - Maximum length of parameter in URL, HTTP POST request or HTTP body.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            length:
                                description:
                                    - Maximum length of parameter in URL, HTTP POST request or HTTP body in bytes (0 to 2147483647).
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    url-param-length:
                        description:
                            - Maximum length of parameter in URL.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            length:
                                description:
                                    - Maximum length of URL parameter in bytes (0 to 2147483647).
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
                    version:
                        description:
                            - Enable/disable HTTP version check.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Enable/disable the constraint.
                                choices:
                                    - enable
                                    - disable
            extended-log:
                description:
                    - Enable/disable extended logging.
                choices:
                    - enable
                    - disable
            external:
                description:
                    - Disable/Enable external HTTP Inspection.
                choices:
                    - disable
                    - enable
            method:
                description:
                    - Method restriction.
                suboptions:
                    default-allowed-methods:
                        description:
                            - Methods.
                        choices:
                            - get
                            - post
                            - put
                            - head
                            - connect
                            - trace
                            - options
                            - delete
                            - others
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
                    method-policy:
                        description:
                            - HTTP method policy.
                        suboptions:
                            address:
                                description:
                                    - Host address. Source firewall.address.name firewall.addrgrp.name.
                            allowed-methods:
                                description:
                                    - Allowed Methods.
                                choices:
                                    - get
                                    - post
                                    - put
                                    - head
                                    - connect
                                    - trace
                                    - options
                                    - delete
                                    - others
                            id:
                                description:
                                    - HTTP method policy ID.
                                required: true
                            pattern:
                                description:
                                    - URL pattern.
                            regex:
                                description:
                                    - Enable/disable regular expression based pattern match.
                                choices:
                                    - enable
                                    - disable
                    severity:
                        description:
                            - Severity.
                        choices:
                            - high
                            - medium
                            - low
                    status:
                        description:
                            - Status.
                        choices:
                            - enable
                            - disable
            name:
                description:
                    - WAF Profile name.
                required: true
            signature:
                description:
                    - WAF signatures.
                suboptions:
                    credit-card-detection-threshold:
                        description:
                            - The minimum number of Credit cards to detect violation.
                    custom-signature:
                        description:
                            - Custom signature.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                                    - erase
                            case-sensitivity:
                                description:
                                    - Case sensitivity in pattern.
                                choices:
                                    - disable
                                    - enable
                            direction:
                                description:
                                    - Traffic direction.
                                choices:
                                    - request
                                    - response
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            name:
                                description:
                                    - Signature name.
                                required: true
                            pattern:
                                description:
                                    - Match pattern.
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Status.
                                choices:
                                    - enable
                                    - disable
                            target:
                                description:
                                    - Match HTTP target.
                                choices:
                                    - arg
                                    - arg-name
                                    - req-body
                                    - req-cookie
                                    - req-cookie-name
                                    - req-filename
                                    - req-header
                                    - req-header-name
                                    - req-raw-uri
                                    - req-uri
                                    - resp-body
                                    - resp-hdr
                                    - resp-status
                    disabled-signature:
                        description:
                            - Disabled signatures
                        suboptions:
                            id:
                                description:
                                    - Signature ID. Source waf.signature.id.
                                required: true
                    disabled-sub-class:
                        description:
                            - Disabled signature subclasses.
                        suboptions:
                            id:
                                description:
                                    - Signature subclass ID. Source waf.sub-class.id.
                                required: true
                    main-class:
                        description:
                            - Main signature class.
                        suboptions:
                            action:
                                description:
                                    - Action.
                                choices:
                                    - allow
                                    - block
                                    - erase
                            id:
                                description:
                                    - Main signature class ID. Source waf.main-class.id.
                                required: true
                            log:
                                description:
                                    - Enable/disable logging.
                                choices:
                                    - enable
                                    - disable
                            severity:
                                description:
                                    - Severity.
                                choices:
                                    - high
                                    - medium
                                    - low
                            status:
                                description:
                                    - Status.
                                choices:
                                    - enable
                                    - disable
            url-access:
                description:
                    - URL access list
                suboptions:
                    access-pattern:
                        description:
                            - URL access pattern.
                        suboptions:
                            id:
                                description:
                                    - URL access pattern ID.
                                required: true
                            negate:
                                description:
                                    - Enable/disable match negation.
                                choices:
                                    - enable
                                    - disable
                            pattern:
                                description:
                                    - URL pattern.
                            regex:
                                description:
                                    - Enable/disable regular expression based pattern match.
                                choices:
                                    - enable
                                    - disable
                            srcaddr:
                                description:
                                    - Source address. Source firewall.address.name firewall.addrgrp.name.
                    action:
                        description:
                            - Action.
                        choices:
                            - bypass
                            - permit
                            - block
                    address:
                        description:
                            - Host address. Source firewall.address.name firewall.addrgrp.name.
                    id:
                        description:
                            - URL access ID.
                        required: true
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
                    severity:
                        description:
                            - Severity.
                        choices:
                            - high
                            - medium
                            - low
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
  tasks:
  - name: Web application firewall configuration.
    fortios_waf_profile:
      host:  "{{ host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{ vdom }}"
      https: "False"
      waf_profile:
        state: "present"
        address-list:
            blocked-address:
             -
                name: "default_name_5 (source firewall.address.name firewall.addrgrp.name)"
            blocked-log: "enable"
            severity: "high"
            status: "enable"
            trusted-address:
             -
                name: "default_name_10 (source firewall.address.name firewall.addrgrp.name)"
        comment: "Comment."
        constraint:
            content-length:
                action: "allow"
                length: "15"
                log: "enable"
                severity: "high"
                status: "enable"
            exception:
             -
                address: "<your_own_value> (source firewall.address.name firewall.addrgrp.name)"
                content-length: "enable"
                header-length: "enable"
                hostname: "enable"
                id:  "24"
                line-length: "enable"
                malformed: "enable"
                max-cookie: "enable"
                max-header-line: "enable"
                max-range-segment: "enable"
                max-url-param: "enable"
                method: "enable"
                param-length: "enable"
                pattern: "<your_own_value>"
                regex: "enable"
                url-param-length: "enable"
                version: "enable"
            header-length:
                action: "allow"
                length: "39"
                log: "enable"
                severity: "high"
                status: "enable"
            hostname:
                action: "allow"
                log: "enable"
                severity: "high"
                status: "enable"
            line-length:
                action: "allow"
                length: "50"
                log: "enable"
                severity: "high"
                status: "enable"
            malformed:
                action: "allow"
                log: "enable"
                severity: "high"
                status: "enable"
            max-cookie:
                action: "allow"
                log: "enable"
                max-cookie: "62"
                severity: "high"
                status: "enable"
            max-header-line:
                action: "allow"
                log: "enable"
                max-header-line: "68"
                severity: "high"
                status: "enable"
            max-range-segment:
                action: "allow"
                log: "enable"
                max-range-segment: "74"
                severity: "high"
                status: "enable"
            max-url-param:
                action: "allow"
                log: "enable"
                max-url-param: "80"
                severity: "high"
                status: "enable"
            method:
                action: "allow"
                log: "enable"
                severity: "high"
                status: "enable"
            param-length:
                action: "allow"
                length: "90"
                log: "enable"
                severity: "high"
                status: "enable"
            url-param-length:
                action: "allow"
                length: "96"
                log: "enable"
                severity: "high"
                status: "enable"
            version:
                action: "allow"
                log: "enable"
                severity: "high"
                status: "enable"
        extended-log: "enable"
        external: "disable"
        method:
            default-allowed-methods: "get"
            log: "enable"
            method-policy:
             -
                address: "<your_own_value> (source firewall.address.name firewall.addrgrp.name)"
                allowed-methods: "get"
                id:  "113"
                pattern: "<your_own_value>"
                regex: "enable"
            severity: "high"
            status: "enable"
        name: "default_name_118"
        signature:
            credit-card-detection-threshold: "120"
            custom-signature:
             -
                action: "allow"
                case-sensitivity: "disable"
                direction: "request"
                log: "enable"
                name: "default_name_126"
                pattern: "<your_own_value>"
                severity: "high"
                status: "enable"
                target: "arg"
            disabled-signature:
             -
                id:  "132 (source waf.signature.id)"
            disabled-sub-class:
             -
                id:  "134 (source waf.sub-class.id)"
            main-class:
             -
                action: "allow"
                id:  "137 (source waf.main-class.id)"
                log: "enable"
                severity: "high"
                status: "enable"
        url-access:
         -
            access-pattern:
             -
                id:  "143"
                negate: "enable"
                pattern: "<your_own_value>"
                regex: "enable"
                srcaddr: "<your_own_value> (source firewall.address.name firewall.addrgrp.name)"
            action: "bypass"
            address: "<your_own_value> (source firewall.address.name firewall.addrgrp.name)"
            id:  "150"
            log: "enable"
            severity: "high"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule


def login(data, fos):
    host = data['host']
    username = data['username']
    password = data['password']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password)


def filter_waf_profile_data(json):
    option_list = ['address-list', 'comment', 'constraint',
                   'extended-log', 'external', 'method',
                   'name', 'signature', 'url-access']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def waf_profile(data, fos):
    vdom = data['vdom']
    waf_profile_data = data['waf_profile']
    filtered_data = filter_waf_profile_data(waf_profile_data)

    if waf_profile_data['state'] == "present":
        return fos.set('waf',
                       'profile',
                       data=filtered_data,
                       vdom=vdom)

    elif waf_profile_data['state'] == "absent":
        return fos.delete('waf',
                          'profile',
                          mkey=filtered_data['name'],
                          vdom=vdom)


def fortios_waf(data, fos):
    login(data, fos)

    if data['waf_profile']:
        resp = waf_profile(data, fos)

    fos.logout()
    return not resp['status'] == "success", resp['status'] == "success", resp


def main():
    fields = {
        "host": {"required": True, "type": "str"},
        "username": {"required": True, "type": "str"},
        "password": {"required": False, "type": "str", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "waf_profile": {
            "required": False, "type": "dict",
            "options": {
                "state": {"required": True, "type": "str",
                          "choices": ["present", "absent"]},
                "address-list": {"required": False, "type": "dict",
                                 "options": {
                                     "blocked-address": {"required": False, "type": "list",
                                                         "options": {
                                                             "name": {"required": True, "type": "str"}
                                                         }},
                                     "blocked-log": {"required": False, "type": "str",
                                                     "choices": ["enable", "disable"]},
                                     "severity": {"required": False, "type": "str",
                                                  "choices": ["high", "medium", "low"]},
                                     "status": {"required": False, "type": "str",
                                                "choices": ["enable", "disable"]},
                                     "trusted-address": {"required": False, "type": "list",
                                                         "options": {
                                                             "name": {"required": True, "type": "str"}
                                                         }}
                                 }},
                "comment": {"required": False, "type": "str"},
                "constraint": {"required": False, "type": "dict",
                               "options": {
                                   "content-length": {"required": False, "type": "dict",
                                                      "options": {
                                                          "action": {"required": False, "type": "str",
                                                                     "choices": ["allow", "block"]},
                                                          "length": {"required": False, "type": "int"},
                                                          "log": {"required": False, "type": "str",
                                                                  "choices": ["enable", "disable"]},
                                                          "severity": {"required": False, "type": "str",
                                                                       "choices": ["high", "medium", "low"]},
                                                          "status": {"required": False, "type": "str",
                                                                     "choices": ["enable", "disable"]}
                                                      }},
                                   "exception": {"required": False, "type": "list",
                                                 "options": {
                                                     "address": {"required": False, "type": "str"},
                                                     "content-length": {"required": False, "type": "str",
                                                                        "choices": ["enable", "disable"]},
                                                     "header-length": {"required": False, "type": "str",
                                                                       "choices": ["enable", "disable"]},
                                                     "hostname": {"required": False, "type": "str",
                                                                  "choices": ["enable", "disable"]},
                                                     "id": {"required": True, "type": "int"},
                                                     "line-length": {"required": False, "type": "str",
                                                                     "choices": ["enable", "disable"]},
                                                     "malformed": {"required": False, "type": "str",
                                                                   "choices": ["enable", "disable"]},
                                                     "max-cookie": {"required": False, "type": "str",
                                                                    "choices": ["enable", "disable"]},
                                                     "max-header-line": {"required": False, "type": "str",
                                                                         "choices": ["enable", "disable"]},
                                                     "max-range-segment": {"required": False, "type": "str",
                                                                           "choices": ["enable", "disable"]},
                                                     "max-url-param": {"required": False, "type": "str",
                                                                       "choices": ["enable", "disable"]},
                                                     "method": {"required": False, "type": "str",
                                                                "choices": ["enable", "disable"]},
                                                     "param-length": {"required": False, "type": "str",
                                                                      "choices": ["enable", "disable"]},
                                                     "pattern": {"required": False, "type": "str"},
                                                     "regex": {"required": False, "type": "str",
                                                               "choices": ["enable", "disable"]},
                                                     "url-param-length": {"required": False, "type": "str",
                                                                          "choices": ["enable", "disable"]},
                                                     "version": {"required": False, "type": "str",
                                                                 "choices": ["enable", "disable"]}
                                                 }},
                                   "header-length": {"required": False, "type": "dict",
                                                     "options": {
                                                         "action": {"required": False, "type": "str",
                                                                    "choices": ["allow", "block"]},
                                                         "length": {"required": False, "type": "int"},
                                                         "log": {"required": False, "type": "str",
                                                                 "choices": ["enable", "disable"]},
                                                         "severity": {"required": False, "type": "str",
                                                                      "choices": ["high", "medium", "low"]},
                                                         "status": {"required": False, "type": "str",
                                                                    "choices": ["enable", "disable"]}
                                                     }},
                                   "hostname": {"required": False, "type": "dict",
                                                "options": {
                                                    "action": {"required": False, "type": "str",
                                                               "choices": ["allow", "block"]},
                                                    "log": {"required": False, "type": "str",
                                                            "choices": ["enable", "disable"]},
                                                    "severity": {"required": False, "type": "str",
                                                                 "choices": ["high", "medium", "low"]},
                                                    "status": {"required": False, "type": "str",
                                                               "choices": ["enable", "disable"]}
                                                }},
                                   "line-length": {"required": False, "type": "dict",
                                                   "options": {
                                                       "action": {"required": False, "type": "str",
                                                                  "choices": ["allow", "block"]},
                                                       "length": {"required": False, "type": "int"},
                                                       "log": {"required": False, "type": "str",
                                                               "choices": ["enable", "disable"]},
                                                       "severity": {"required": False, "type": "str",
                                                                    "choices": ["high", "medium", "low"]},
                                                       "status": {"required": False, "type": "str",
                                                                  "choices": ["enable", "disable"]}
                                                   }},
                                   "malformed": {"required": False, "type": "dict",
                                                 "options": {
                                                     "action": {"required": False, "type": "str",
                                                                "choices": ["allow", "block"]},
                                                     "log": {"required": False, "type": "str",
                                                             "choices": ["enable", "disable"]},
                                                     "severity": {"required": False, "type": "str",
                                                                  "choices": ["high", "medium", "low"]},
                                                     "status": {"required": False, "type": "str",
                                                                "choices": ["enable", "disable"]}
                                                 }},
                                   "max-cookie": {"required": False, "type": "dict",
                                                  "options": {
                                                      "action": {"required": False, "type": "str",
                                                                 "choices": ["allow", "block"]},
                                                      "log": {"required": False, "type": "str",
                                                              "choices": ["enable", "disable"]},
                                                      "max-cookie": {"required": False, "type": "int"},
                                                      "severity": {"required": False, "type": "str",
                                                                   "choices": ["high", "medium", "low"]},
                                                      "status": {"required": False, "type": "str",
                                                                 "choices": ["enable", "disable"]}
                                                  }},
                                   "max-header-line": {"required": False, "type": "dict",
                                                       "options": {
                                                           "action": {"required": False, "type": "str",
                                                                      "choices": ["allow", "block"]},
                                                           "log": {"required": False, "type": "str",
                                                                   "choices": ["enable", "disable"]},
                                                           "max-header-line": {"required": False, "type": "int"},
                                                           "severity": {"required": False, "type": "str",
                                                                        "choices": ["high", "medium", "low"]},
                                                           "status": {"required": False, "type": "str",
                                                                      "choices": ["enable", "disable"]}
                                                       }},
                                   "max-range-segment": {"required": False, "type": "dict",
                                                         "options": {
                                                             "action": {"required": False, "type": "str",
                                                                        "choices": ["allow", "block"]},
                                                             "log": {"required": False, "type": "str",
                                                                     "choices": ["enable", "disable"]},
                                                             "max-range-segment": {"required": False, "type": "int"},
                                                             "severity": {"required": False, "type": "str",
                                                                          "choices": ["high", "medium", "low"]},
                                                             "status": {"required": False, "type": "str",
                                                                        "choices": ["enable", "disable"]}
                                                         }},
                                   "max-url-param": {"required": False, "type": "dict",
                                                     "options": {
                                                         "action": {"required": False, "type": "str",
                                                                    "choices": ["allow", "block"]},
                                                         "log": {"required": False, "type": "str",
                                                                 "choices": ["enable", "disable"]},
                                                         "max-url-param": {"required": False, "type": "int"},
                                                         "severity": {"required": False, "type": "str",
                                                                      "choices": ["high", "medium", "low"]},
                                                         "status": {"required": False, "type": "str",
                                                                    "choices": ["enable", "disable"]}
                                                     }},
                                   "method": {"required": False, "type": "dict",
                                              "options": {
                                                  "action": {"required": False, "type": "str",
                                                             "choices": ["allow", "block"]},
                                                  "log": {"required": False, "type": "str",
                                                          "choices": ["enable", "disable"]},
                                                  "severity": {"required": False, "type": "str",
                                                               "choices": ["high", "medium", "low"]},
                                                  "status": {"required": False, "type": "str",
                                                             "choices": ["enable", "disable"]}
                                              }},
                                   "param-length": {"required": False, "type": "dict",
                                                    "options": {
                                                        "action": {"required": False, "type": "str",
                                                                   "choices": ["allow", "block"]},
                                                        "length": {"required": False, "type": "int"},
                                                        "log": {"required": False, "type": "str",
                                                                "choices": ["enable", "disable"]},
                                                        "severity": {"required": False, "type": "str",
                                                                     "choices": ["high", "medium", "low"]},
                                                        "status": {"required": False, "type": "str",
                                                                   "choices": ["enable", "disable"]}
                                                    }},
                                   "url-param-length": {"required": False, "type": "dict",
                                                        "options": {
                                                            "action": {"required": False, "type": "str",
                                                                       "choices": ["allow", "block"]},
                                                            "length": {"required": False, "type": "int"},
                                                            "log": {"required": False, "type": "str",
                                                                    "choices": ["enable", "disable"]},
                                                            "severity": {"required": False, "type": "str",
                                                                         "choices": ["high", "medium", "low"]},
                                                            "status": {"required": False, "type": "str",
                                                                       "choices": ["enable", "disable"]}
                                                        }},
                                   "version": {"required": False, "type": "dict",
                                               "options": {
                                                   "action": {"required": False, "type": "str",
                                                              "choices": ["allow", "block"]},
                                                   "log": {"required": False, "type": "str",
                                                           "choices": ["enable", "disable"]},
                                                   "severity": {"required": False, "type": "str",
                                                                "choices": ["high", "medium", "low"]},
                                                   "status": {"required": False, "type": "str",
                                                              "choices": ["enable", "disable"]}
                                               }}
                               }},
                "extended-log": {"required": False, "type": "str",
                                 "choices": ["enable", "disable"]},
                "external": {"required": False, "type": "str",
                             "choices": ["disable", "enable"]},
                "method": {"required": False, "type": "dict",
                           "options": {
                               "default-allowed-methods": {"required": False, "type": "str",
                                                           "choices": ["get", "post", "put",
                                                                       "head", "connect", "trace",
                                                                       "options", "delete", "others"]},
                               "log": {"required": False, "type": "str",
                                       "choices": ["enable", "disable"]},
                               "method-policy": {"required": False, "type": "list",
                                                 "options": {
                                                     "address": {"required": False, "type": "str"},
                                                     "allowed-methods": {"required": False, "type": "str",
                                                                         "choices": ["get", "post", "put",
                                                                                     "head", "connect", "trace",
                                                                                     "options", "delete", "others"]},
                                                     "id": {"required": True, "type": "int"},
                                                     "pattern": {"required": False, "type": "str"},
                                                     "regex": {"required": False, "type": "str",
                                                               "choices": ["enable", "disable"]}
                                                 }},
                               "severity": {"required": False, "type": "str",
                                            "choices": ["high", "medium", "low"]},
                               "status": {"required": False, "type": "str",
                                          "choices": ["enable", "disable"]}
                           }},
                "name": {"required": True, "type": "str"},
                "signature": {"required": False, "type": "dict",
                              "options": {
                                  "credit-card-detection-threshold": {"required": False, "type": "int"},
                                  "custom-signature": {"required": False, "type": "list",
                                                       "options": {
                                                           "action": {"required": False, "type": "str",
                                                                      "choices": ["allow", "block", "erase"]},
                                                           "case-sensitivity": {"required": False, "type": "str",
                                                                                "choices": ["disable", "enable"]},
                                                           "direction": {"required": False, "type": "str",
                                                                         "choices": ["request", "response"]},
                                                           "log": {"required": False, "type": "str",
                                                                   "choices": ["enable", "disable"]},
                                                           "name": {"required": True, "type": "str"},
                                                           "pattern": {"required": False, "type": "str"},
                                                           "severity": {"required": False, "type": "str",
                                                                        "choices": ["high", "medium", "low"]},
                                                           "status": {"required": False, "type": "str",
                                                                      "choices": ["enable", "disable"]},
                                                           "target": {"required": False, "type": "str",
                                                                      "choices": ["arg", "arg-name", "req-body",
                                                                                  "req-cookie", "req-cookie-name", "req-filename",
                                                                                  "req-header", "req-header-name", "req-raw-uri",
                                                                                  "req-uri", "resp-body", "resp-hdr",
                                                                                  "resp-status"]}
                                                       }},
                                  "disabled-signature": {"required": False, "type": "list",
                                                         "options": {
                                                             "id": {"required": True, "type": "int"}
                                                         }},
                                  "disabled-sub-class": {"required": False, "type": "list",
                                                         "options": {
                                                             "id": {"required": True, "type": "int"}
                                                         }},
                                  "main-class": {"required": False, "type": "list",
                                                 "options": {
                                                     "action": {"required": False, "type": "str",
                                                                "choices": ["allow", "block", "erase"]},
                                                     "id": {"required": True, "type": "int"},
                                                     "log": {"required": False, "type": "str",
                                                             "choices": ["enable", "disable"]},
                                                     "severity": {"required": False, "type": "str",
                                                                  "choices": ["high", "medium", "low"]},
                                                     "status": {"required": False, "type": "str",
                                                                "choices": ["enable", "disable"]}
                                                 }}
                              }},
                "url-access": {"required": False, "type": "list",
                               "options": {
                                   "access-pattern": {"required": False, "type": "list",
                                                      "options": {
                                                          "id": {"required": True, "type": "int"},
                                                          "negate": {"required": False, "type": "str",
                                                                     "choices": ["enable", "disable"]},
                                                          "pattern": {"required": False, "type": "str"},
                                                          "regex": {"required": False, "type": "str",
                                                                    "choices": ["enable", "disable"]},
                                                          "srcaddr": {"required": False, "type": "str"}
                                                      }},
                                   "action": {"required": False, "type": "str",
                                              "choices": ["bypass", "permit", "block"]},
                                   "address": {"required": False, "type": "str"},
                                   "id": {"required": True, "type": "int"},
                                   "log": {"required": False, "type": "str",
                                           "choices": ["enable", "disable"]},
                                   "severity": {"required": False, "type": "str",
                                                "choices": ["high", "medium", "low"]}
                               }}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)
    try:
        from fortiosapi import FortiOSAPI
    except ImportError:
        module.fail_json(msg="fortiosapi module is required")

    fos = FortiOSAPI()

    is_error, has_changed, result = fortios_waf(module.params, fos)

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from units.compat.mock import patch
from ansible.modules.network.onyx import onyx_ptp_global
from units.modules.utils import set_module_args
from .onyx_module import TestOnyxModule, load_fixture


class TestOnyxPtpModule(TestOnyxModule):

    module = onyx_ptp_global

    def setUp(self):
        self._ptp_enabled = True
        self._ntp_enabled = True
        super(TestOnyxPtpModule, self).setUp()

        self.mock_get_ptp_config = patch.object(onyx_ptp_global.OnyxPtpGlobalModule, "_show_ptp_config")
        self.get_ptp_config = self.mock_get_ptp_config.start()
        self.mock_get_ntp_config = patch.object(onyx_ptp_global.OnyxPtpGlobalModule, "_show_ntp_config")
        self.get_ntp_config = self.mock_get_ntp_config.start()

        self.mock_load_config = patch(
            'ansible.module_utils.network.onyx.onyx.load_config')
        self.load_config = self.mock_load_config.start()

    def tearDown(self):
        super(TestOnyxPtpModule, self).tearDown()
        self.mock_get_ptp_config.stop()
        self.mock_get_ntp_config.stop()
        self.mock_load_config.stop()

    def load_fixtures(self, commands=None, transport='cli'):
        if self._ptp_enabled:
            config_file = 'onyx_show_ptp_clock.cfg'
            self.get_ptp_config.return_value = load_fixture(config_file)
        else:
            self.get_ptp_config.return_value = None

        config_file = 'onyx_show_ntp_configured.cfg'
        ret_val = load_fixture(config_file)
        if self._ntp_enabled:
            ret_val[0]['NTP enabled'] = 'yes'
        self.get_ntp_config.return_value = ret_val
        self.load_config.return_value = None

    def test_ptp_enabled_no_change(self):
        set_module_args(dict(ptp_state='enabled'))
        self.execute_module(changed=False)

    def test_ptp_enabled_with_change(self):
        self._ptp_enabled = False
        set_module_args(dict(ptp_state='enabled'))
        commands = ['protocol ptp']
        self.execute_module(changed=True, commands=commands)

    def test_ptp_disabled_no_change(self):
        self._ptp_enabled = False
        set_module_args(dict(ptp_state='disabled'))
        self.execute_module(changed=False)

    def test_ptp_disabled_with_change(self):
        set_module_args(dict(ptp_state='disabled'))
        commands = ['no protocol ptp']
        self.execute_module(changed=True, commands=commands)

    def test_ntp_enabled_no_change(self):
        self._ptp_enabled = False
        set_module_args(dict(ntp_state='enabled',
                             ptp_state='disabled'))
        self.execute_module(changed=False)

    def test_ntp_enabled_with_change(self):
        self._ptp_enabled = False
        self._ntp_enabled = False
        set_module_args(dict(ntp_state='enabled',
                             ptp_state='disabled'))
        commands = ['ntp enable']
        self.execute_module(changed=True, commands=commands)

    def test_ntp_disabled_no_change(self):
        self._ntp_enabled = False
        set_module_args(dict(ntp_state='disabled'))
        self.execute_module(changed=False)

    def test_ntp_disabled_with_change(self):
        set_module_args(dict(ntp_state='disabled'))
        commands = ['no ntp enable']
        self.execute_module(changed=True, commands=commands)

    def test_set_domain_no_change(self):
        self._ntp_enabled = False
        set_module_args(dict(ntp_state='disabled',
                             domain=127))
        self.execute_module(changed=False)

    def test_set_domain_with_change(self):
        set_module_args(dict(domain=100))
        commands = ['ptp domain 100']
        self.execute_module(changed=True, commands=commands)

    def test_set_primary_priority_no_change(self):
        set_module_args(dict(primary_priority=128))
        self.execute_module(changed=False)

    def test_set_primary_priority_with_change(self):
        set_module_args(dict(primary_priority=250))
        commands = ['ptp priority1 250']
        self.execute_module(changed=True, commands=commands)

    def test_set_secondary_priority_no_change(self):
        set_module_args(dict(secondary_priority=128))
        self.execute_module(changed=False)

    def test_set_secondary_priority_with_change(self):
        set_module_args(dict(secondary_priority=190))
        commands = ['ptp priority2 190']
        self.execute_module(changed=True, commands=commands)



#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# the lib use python logging can get it if the following is set in your
# Ansible config.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_spamfilter_profile
short_description: Configure AntiSpam profiles in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS by allowing the
      user to set and modify spamfilter feature and profile category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.2
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
       description:
            - FortiOS or FortiGate ip address.
       required: true
    username:
        description:
            - FortiOS or FortiGate username.
        required: true
    password:
        description:
            - FortiOS or FortiGate password.
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS
              protocol
        type: bool
        default: true
    spamfilter_profile:
        description:
            - Configure AntiSpam profiles.
        default: null
        suboptions:
            state:
                description:
                    - Indicates whether to create or remove the object
                choices:
                    - present
                    - absent
            comment:
                description:
                    - Comment.
            external:
                description:
                    - Enable/disable external Email inspection.
                choices:
                    - enable
                    - disable
            flow-based:
                description:
                    - Enable/disable flow-based spam filtering.
                choices:
                    - enable
                    - disable
            gmail:
                description:
                    - Gmail.
                suboptions:
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
            imap:
                description:
                    - IMAP.
                suboptions:
                    action:
                        description:
                            - Action for spam email.
                        choices:
                            - pass
                            - tag
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
                    tag-msg:
                        description:
                            - Subject text or header added to spam email.
                    tag-type:
                        description:
                            - Tag subject or header for spam email.
                        choices:
                            - subject
                            - header
                            - spaminfo
            mapi:
                description:
                    - MAPI.
                suboptions:
                    action:
                        description:
                            - Action for spam email.
                        choices:
                            - pass
                            - discard
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
            msn-hotmail:
                description:
                    - MSN Hotmail.
                suboptions:
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
            name:
                description:
                    - Profile name.
                required: true
            options:
                description:
                    - Options.
                choices:
                    - bannedword
                    - spambwl
                    - spamfsip
                    - spamfssubmit
                    - spamfschksum
                    - spamfsurl
                    - spamhelodns
                    - spamraddrdns
                    - spamrbl
                    - spamhdrcheck
                    - spamfsphish
            pop3:
                description:
                    - POP3.
                suboptions:
                    action:
                        description:
                            - Action for spam email.
                        choices:
                            - pass
                            - tag
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
                    tag-msg:
                        description:
                            - Subject text or header added to spam email.
                    tag-type:
                        description:
                            - Tag subject or header for spam email.
                        choices:
                            - subject
                            - header
                            - spaminfo
            replacemsg-group:
                description:
                    - Replacement message group. Source system.replacemsg-group.name.
            smtp:
                description:
                    - SMTP.
                suboptions:
                    action:
                        description:
                            - Action for spam email.
                        choices:
                            - pass
                            - tag
                            - discard
                    hdrip:
                        description:
                            - Enable/disable SMTP email header IP checks for spamfsip, spamrbl and spambwl filters.
                        choices:
                            - disable
                            - enable
                    local-override:
                        description:
                            - Enable/disable local filter to override SMTP remote check result.
                        choices:
                            - disable
                            - enable
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
                    tag-msg:
                        description:
                            - Subject text or header added to spam email.
                    tag-type:
                        description:
                            - Tag subject or header for spam email.
                        choices:
                            - subject
                            - header
                            - spaminfo
            spam-bwl-table:
                description:
                    - Anti-spam black/white list table ID. Source spamfilter.bwl.id.
            spam-bword-table:
                description:
                    - Anti-spam banned word table ID. Source spamfilter.bword.id.
            spam-bword-threshold:
                description:
                    - Spam banned word threshold.
            spam-filtering:
                description:
                    - Enable/disable spam filtering.
                choices:
                    - enable
                    - disable
            spam-iptrust-table:
                description:
                    - Anti-spam IP trust table ID. Source spamfilter.iptrust.id.
            spam-log:
                description:
                    - Enable/disable spam logging for email filtering.
                choices:
                    - disable
                    - enable
            spam-log-fortiguard-response:
                description:
                    - Enable/disable logging FortiGuard spam response.
                choices:
                    - disable
                    - enable
            spam-mheader-table:
                description:
                    - Anti-spam MIME header table ID. Source spamfilter.mheader.id.
            spam-rbl-table:
                description:
                    - Anti-spam DNSBL table ID. Source spamfilter.dnsbl.id.
            yahoo-mail:
                description:
                    - Yahoo! Mail.
                suboptions:
                    log:
                        description:
                            - Enable/disable logging.
                        choices:
                            - enable
                            - disable
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
  tasks:
  - name: Configure AntiSpam profiles.
    fortios_spamfilter_profile:
      host:  "{{ host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{ vdom }}"
      https: "False"
      spamfilter_profile:
        state: "present"
        comment: "Comment."
        external: "enable"
        flow-based: "enable"
        gmail:
            log: "enable"
        imap:
            action: "pass"
            log: "enable"
            tag-msg: "<your_own_value>"
            tag-type: "subject"
        mapi:
            action: "pass"
            log: "enable"
        msn-hotmail:
            log: "enable"
        name: "default_name_18"
        options: "bannedword"
        pop3:
            action: "pass"
            log: "enable"
            tag-msg: "<your_own_value>"
            tag-type: "subject"
        replacemsg-group: "<your_own_value> (source system.replacemsg-group.name)"
        smtp:
            action: "pass"
            hdrip: "disable"
            local-override: "disable"
            log: "enable"
            tag-msg: "<your_own_value>"
            tag-type: "subject"
        spam-bwl-table: "33 (source spamfilter.bwl.id)"
        spam-bword-table: "34 (source spamfilter.bword.id)"
        spam-bword-threshold: "35"
        spam-filtering: "enable"
        spam-iptrust-table: "37 (source spamfilter.iptrust.id)"
        spam-log: "disable"
        spam-log-fortiguard-response: "disable"
        spam-mheader-table: "40 (source spamfilter.mheader.id)"
        spam-rbl-table: "41 (source spamfilter.dnsbl.id)"
        yahoo-mail:
            log: "enable"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule

fos = None


def login(data):
    host = data['host']
    username = data['username']
    password = data['password']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password)


def filter_spamfilter_profile_data(json):
    option_list = ['comment', 'external', 'flow-based',
                   'gmail', 'imap', 'mapi',
                   'msn-hotmail', 'name', 'options',
                   'pop3', 'replacemsg-group', 'smtp',
                   'spam-bwl-table', 'spam-bword-table', 'spam-bword-threshold',
                   'spam-filtering', 'spam-iptrust-table', 'spam-log',
                   'spam-log-fortiguard-response', 'spam-mheader-table', 'spam-rbl-table',
                   'yahoo-mail']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def flatten_multilists_attributes(data):
    multilist_attrs = [[u'options'], [u'imap', u'tag-type'], [u'pop3', u'tag-type'], [u'smtp', u'tag-type']]

    for attr in multilist_attrs:
        try:
            path = "data['" + "']['".join(elem for elem in attr) + "']"
            current_val = eval(path)
            flattened_val = ' '.join(elem for elem in current_val)
            exec(path + '= flattened_val')
        except BaseException:
            pass

    return data


def spamfilter_profile(data, fos):
    vdom = data['vdom']
    spamfilter_profile_data = data['spamfilter_profile']
    flattened_data = flatten_multilists_attributes(spamfilter_profile_data)
    filtered_data = filter_spamfilter_profile_data(flattened_data)
    if spamfilter_profile_data['state'] == "present":
        return fos.set('spamfilter',
                       'profile',
                       data=filtered_data,
                       vdom=vdom)

    elif spamfilter_profile_data['state'] == "absent":
        return fos.delete('spamfilter',
                          'profile',
                          mkey=filtered_data['name'],
                          vdom=vdom)


def fortios_spamfilter(data, fos):
    login(data)

    if data['spamfilter_profile']:
        resp = spamfilter_profile(data, fos)

    fos.logout()
    return not resp['status'] == "success", resp['status'] == "success", resp


def main():
    fields = {
        "host": {"required": True, "type": "str"},
        "username": {"required": True, "type": "str"},
        "password": {"required": False, "type": "str", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "spamfilter_profile": {
            "required": False, "type": "dict",
            "options": {
                "state": {"required": True, "type": "str",
                          "choices": ["present", "absent"]},
                "comment": {"required": False, "type": "str"},
                "external": {"required": False, "type": "str",
                             "choices": ["enable", "disable"]},
                "flow-based": {"required": False, "type": "str",
                               "choices": ["enable", "disable"]},
                "gmail": {"required": False, "type": "dict",
                          "options": {
                              "log": {"required": False, "type": "str",
                                      "choices": ["enable", "disable"]}
                          }},
                "imap": {"required": False, "type": "dict",
                         "options": {
                             "action": {"required": False, "type": "str",
                                        "choices": ["pass", "tag"]},
                             "log": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                             "tag-msg": {"required": False, "type": "str"},
                             "tag-type": {"required": False, "type": "list",
                                          "choices": ["subject", "header", "spaminfo"]}
                         }},
                "mapi": {"required": False, "type": "dict",
                         "options": {
                             "action": {"required": False, "type": "str",
                                        "choices": ["pass", "discard"]},
                             "log": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]}
                         }},
                "msn-hotmail": {"required": False, "type": "dict",
                                "options": {
                                    "log": {"required": False, "type": "str",
                                            "choices": ["enable", "disable"]}
                                }},
                "name": {"required": True, "type": "str"},
                "options": {"required": False, "type": "list",
                            "choices": ["bannedword", "spambwl", "spamfsip",
                                        "spamfssubmit", "spamfschksum", "spamfsurl",
                                        "spamhelodns", "spamraddrdns", "spamrbl",
                                        "spamhdrcheck", "spamfsphish"]},
                "pop3": {"required": False, "type": "dict",
                         "options": {
                             "action": {"required": False, "type": "str",
                                        "choices": ["pass", "tag"]},
                             "log": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                             "tag-msg": {"required": False, "type": "str"},
                             "tag-type": {"required": False, "type": "list",
                                          "choices": ["subject", "header", "spaminfo"]}
                         }},
                "replacemsg-group": {"required": False, "type": "str"},
                "smtp": {"required": False, "type": "dict",
                         "options": {
                             "action": {"required": False, "type": "str",
                                        "choices": ["pass", "tag", "discard"]},
                             "hdrip": {"required": False, "type": "str",
                                       "choices": ["disable", "enable"]},
                             "local-override": {"required": False, "type": "str",
                                                "choices": ["disable", "enable"]},
                             "log": {"required": False, "type": "str",
                                     "choices": ["enable", "disable"]},
                             "tag-msg": {"required": False, "type": "str"},
                             "tag-type": {"required": False, "type": "list",
                                          "choices": ["subject", "header", "spaminfo"]}
                         }},
                "spam-bwl-table": {"required": False, "type": "int"},
                "spam-bword-table": {"required": False, "type": "int"},
                "spam-bword-threshold": {"required": False, "type": "int"},
                "spam-filtering": {"required": False, "type": "str",
                                   "choices": ["enable", "disable"]},
                "spam-iptrust-table": {"required": False, "type": "int"},
                "spam-log": {"required": False, "type": "str",
                             "choices": ["disable", "enable"]},
                "spam-log-fortiguard-response": {"required": False, "type": "str",
                                                 "choices": ["disable", "enable"]},
                "spam-mheader-table": {"required": False, "type": "int"},
                "spam-rbl-table": {"required": False, "type": "int"},
                "yahoo-mail": {"required": False, "type": "dict",
                               "options": {
                                   "log": {"required": False, "type": "str",
                                           "choices": ["enable", "disable"]}
                               }}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)
    try:
        from fortiosapi import FortiOSAPI
    except ImportError:
        module.fail_json(msg="fortiosapi module is required")

    global fos
    fos = FortiOSAPI()

    is_error, has_changed, result = fortios_spamfilter(module.params, fos)

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()
